from node.tests import NodeTestCase
from yafowil.tsf import DummyTranslationStringFactory


class TestTsf(NodeTestCase):

    def test_tsf(self):
        # Test dummy translation string factory
        _ = DummyTranslationStringFactory('yafowil')
        self.assertEqual(_.domain, 'yafowil')
        self.assertEqual(_('foo'), 'foo')
        self.assertEqual(_('bar', default=u'Bar'), 'Bar')
        self.assertEqual(
            _('baz', default=u'Baz ${bam}', mapping={'bam': 42}),
            'Baz 42'
        )



class Controller(object):
    """Form controller.
    """

    def __init__(self, widget, request):
        """Initialize controller

        ``widget``
            yafowil.base.Widget tree.

        ``request``
            native request
        """
        self.widget = widget
        self.performed = False
        self.error = False
        self.next = None
        self.data = self.widget.extract(request)
        self.request = self.data.request
        self._error(self.data)
        for action in self.actions:
            if not self.triggered(action):
                continue
            if action.attrs.get('skip'):
                if action.attrs.get('next'):
                    self.next = action.attrs['next'](self.request)
                return
            self.performed = True
            if self.error:
                return
            if action.attrs.get('handler'):
                action.attrs['handler'](self.widget, self.data)
            if action.attrs.get('next'):
                self.next = action.attrs['next'](self.request)

    @property
    def rendered(self):
        if not self.performed:
            return self.widget(request=self.request)
        return self.widget(data=self.data)

    @property
    def actions(self):
        result = []

        def collect_actions(level):
            for widget in level.values():
                if widget.attrs.get('action'):
                    result.append(widget)
                collect_actions(widget)

        collect_actions(self.widget)
        return result

    def triggered(self, action):
        action_name = 'action.{0}'.format(action.dottedpath)
        return self.request.get(action_name) is not None

    def _error(self, data):
        if data.errors:
            self.error = True
            return
        for subdata in data.values():
            self._error(subdata)



# -*- coding: utf-8 -*-
from yafowil.base import factory


DOC_PLAIN_TEXT = """\
Plain Text
----------

Simple plaintext widget.

.. code-block:: python

    plaintext = factory('#field:text', props={
        'label': 'Plain Text Input',
        'required': 'Input is required',
        'help': 'Simple Plain Text Field',
    })
"""


def plaintext():
    comp = factory('compound', name='yafowil-plaintext')
    comp['plain'] = factory('#field:text', props={
        'label': 'Plain Text Input',
        'required': 'Input is required',
        'help': 'Simple Plain Text Field',
    })
    return {
        'widget': comp,
        'doc': DOC_PLAIN_TEXT,
        'title': 'Plain Text',
    }


DOC_EMAIL = """
Email
-----

Text input for email addresses.

.. code-block:: python

    email = factory('#field:email', props={
        'label': 'E-Mail Input',
        'help': 'E-Mail Address Field',
    })
"""


def email():
    comp = factory('compound', name='yafowil-email')
    comp['email'] = factory('#field:email', props={
        'label': 'E-Mail Input',
        'help': 'E-Mail Address Field',
    })
    return {
        'widget': comp,
        'doc': DOC_EMAIL,
        'title': 'Email',
    }


DOC_NUMBER = """
Number
------

Input field accepting floating point numbers.

.. code-block:: python

    number = factory('#field:number', props={
        'label': 'Number Input (float)',
        'help': 'Field for floating point number input',
    })
"""


def number():
    comp = factory('compound', name='yafowil-number')
    comp['number'] = factory('#field:number', props={
        'label': 'Number Input (float)',
        'help': 'Field for floating point number input',
    })
    return {
        'widget': comp,
        'doc': DOC_NUMBER,
        'title': 'Number',
    }


DOC_INT = """
Integer
-------

Input field accepting integer values.

.. code-block:: python

    integer = factory('#field:number', props={
        'label': 'Number Input (int)',
        'datatype': 'integer',
        'help': 'Field for integer input',
    })
"""


def integer():
    comp = factory('compound', name='yafowil-integer')
    comp['intnumber'] = factory('#field:number', props={
        'label': 'Number Input (int)',
        'datatype': 'integer',
        'help': 'Field for integer input',
    })
    return {
        'widget': comp,
        'doc': DOC_INT,
        'title': 'Integer',
    }


DOC_PASSWORD = """
Password
--------

Password field.

.. code-block:: python

    password = factory('#field:password', props={
        'label': 'Password Input',
        'help': 'Field for password',
    })
"""


def password():
    comp = factory('compound', name='yafowil-password')
    comp['password'] = factory('#field:password', props={
        'label': 'Password Input',
        'help': 'Field for password',
    })
    return {
        'widget': comp,
        'doc': DOC_PASSWORD,
        'title': 'Password',
    }


DOC_URL = """
URL
---

URL field.

.. code-block:: python

    password = factory('#field:url', props={
        'label': 'URL Input',
        'help': 'Field for URL',
    })
"""


def url():
    comp = factory('compound', name='yafowil-url')
    comp['password'] = factory('#field:url', props={
        'label': 'URL Input',
        'help': 'Field for URL',
    })
    return {
        'widget': comp,
        'doc': DOC_URL,
        'title': 'URL',
    }


DOC_TEXTAREA = """
Textarea
--------

Textarea field.

.. code-block:: python

    textarea = factory('#field:textarea', props={
        'label': 'Textarea',
        'help': 'Textarea field',
        'rows': 5,
    })
"""


def textarea():
    comp = factory('compound', name='yafowil-textarea')
    comp['textarea'] = factory('#field:textarea', props={
        'label': 'Textarea',
        'help': 'Textarea field',
        'rows': 5,
    })
    return {
        'widget': comp,
        'doc': DOC_TEXTAREA,
        'title': 'Textarea',
    }


DOC_RADIO = """\
Radio buttons
-------------

Selection with radio buttons.

.. code-block:: python

    radiobuttons = factory('#field:select', props={
        'label': 'Select',
        'help': 'Single selection as radio buttons',
        'vocabulary': ['Python 2', 'Python 3'],
        'format': 'radio',
    })
"""


def radio():
    comp = factory('compound', name='yafowil-radio')
    comp['radio'] = factory('#field:select', props={
        'label': 'Select',
        'help': 'Single selection as radio buttons',
        'vocabulary': ['Python 2', 'Python 3'],
        'format': 'radio',
    })
    return {
        'widget': comp,
        'doc': DOC_RADIO,
        'title': 'Radio buttons',
    }


DOC_DROPDOWN = """
Selection Dropdown
------------------

Single selection dropdown.

.. code-block:: python

    dropdown = factory('#field:select', props={
        'label': 'Select',
        'help': 'Single selection as dropdown',
        'vocabulary': ['Python', 'Java', 'Perl', 'Erlang', 'C', 'C++', 'C#'],
        'format': 'block',
    })
"""


def dropdown():
    comp = factory('compound', name='yafowil-dropdown')
    comp['dropdown'] = factory('#field:select', props={
        'label': 'Select',
        'help': 'Single selection as dropdown',
        'vocabulary': ['Python', 'Java', 'Perl', 'Erlang', 'C', 'C++', 'C#'],
        'format': 'block',
    })
    return {
        'widget': comp,
        'doc': DOC_DROPDOWN,
        'title': 'Selection Dropdown',
    }


DOC_CHECKBOX = """
Checkbox selection
------------------

Multi selection with checkboxes.

.. code-block:: python

    checkboxes = factory('#field:select', props={
        'label': 'Select',
        'help': 'Multiple selection with checkboxes',
        'vocabulary': ['Python 2', 'Python 3'],
        'format': 'checkbox',
        'multivalued': True,
    })
"""


def checkbox():
    comp = factory('compound', name='yafowil-checkbox')
    comp['checkbox'] = factory('#field:select', props={
        'label': 'Select',
        'help': 'Multiple selection with checkboxes',
        'vocabulary': ['Python 2', 'Python 3'],
        'format': 'checkbox',
        'multivalued': True,
    })
    return {
        'widget': comp,
        'doc': DOC_CHECKBOX,
        'title': 'Checkbox selection',
    }


DOC_BLOCK = """
Selection block
---------------

Multi selection as block.

.. code-block:: python

    selection = factory('#field:select', props={
        'label': 'Select',
        'help': 'multiple as block',
        'vocabulary': ['Python', 'Java', 'Perl', 'Erlang', 'C', 'C++', 'C#'],
        'format': 'block',
        'multivalued': True,
    })
"""


def block():
    comp = factory('compound', name='yafowil-block')
    comp['block'] = factory('#field:select', props={
        'label': 'Select',
        'help': 'multiple as block',
        'vocabulary': ['Python', 'Java', 'Perl', 'Erlang', 'C', 'C++', 'C#'],
        'format': 'block',
        'multivalued': True,
    })
    return {
        'widget': comp,
        'doc': DOC_BLOCK,
        'title': 'Selection block',
    }


DOC_SINGLE_CHECKBOX = """
Single checkbox
---------------

Single checkbox (return True or False). Usually used without surrounding field.
This is an example how to render it for Bootstrap 5

.. code-block:: python

    checkbox = factory('div:help:label:checkbox', props={
        'label': 'Single checkbox',
        'help': 'A single checkbox without vocabulary',
        'div.class': 'form-check',
        'label.class': 'form-check-label',
        'label.position': 'after',
    })
"""


def single_checkbox():
    comp = factory('compound', name='yafowil-single-checkbox')
    comp['checkbox'] = factory('div:help:label:checkbox', props={
        'label': 'Single checkbox',
        'help': 'A single checkbox without vocabulary',
        'div.class': 'form-check',
        'label.class': 'form-check-label',
        'label.position': 'after',
    })
    return {
        'widget': comp,
        'doc': DOC_SINGLE_CHECKBOX,
        'title': 'Single checkbox',
    }


DOC_FILEUPLOAD = """
File upload
-----------

Upload a file

.. code-block:: python

    fileupload = factory('#field:file', props={
        'label': 'File',
        'help': 'Upload a single file',
    })
"""


def fileupload():
    fileupload = factory('compound', name='yafowil-fileupload')
    fileupload['file'] = factory('#field:file', props={
        'label': 'File',
        'help': 'Upload a single file',
    })
    return {
        'widget': fileupload,
        'doc': DOC_FILEUPLOAD,
        'title': 'File upload',
    }


def get_example():
    return [
        plaintext(), email(), number(), integer(), password(), url(),
        textarea(), radio(), dropdown(), checkbox(), block(),
        single_checkbox(), fileupload()
    ]



# -*- coding: utf-8 -*-
from node.behaviors import Attributes
from node.behaviors import DictStorage
from node.behaviors import MappingAdopt
from node.behaviors import MappingConstraints
from node.behaviors import MappingNode
from node.behaviors import NodeAttributes
from node.behaviors import OdictStorage
from node.behaviors import Order
from node.utils import UNSET
from node.utils import instance_property
from plumber import plumbing
from threading import RLock
from yafowil.compat import ITER_TYPES
from yafowil.compat import STR_TYPE
from yafowil.utils import Tag
from yafowil.utils import attr_value
import copy
import webresource as wr


def _dict__repr__(self):
    return '{{{0}}}'.format(
        ', '.join(['{0}: {1}'.format(
            repr(k),
            repr(v)
        ) for k, v in self.items()])
    )


class RuntimeDataAttributes(NodeAttributes):
    __str__ = __repr__ = _dict__repr__


@plumbing(
    Attributes,
    MappingConstraints,
    MappingAdopt,
    MappingNode,
    OdictStorage)
class RuntimeData(object):
    """Holds Runtime data of widget.
    """
    attributes_factory = RuntimeDataAttributes

    def __init__(self,
                 name=None,
                 parent=None,
                 request=UNSET,
                 persist=None,
                 persist_target=None,
                 persist_writer=None):
        self.__name__ = name
        self.__parent__ = parent
        if parent is not None:
            parent[self.name] = self
        self.request = request
        self.value = UNSET
        self.preprocessed = False
        self.extracted = UNSET
        self.rendered = UNSET
        self.errors = list()
        self.translate_callable = lambda msg: msg
        self._persist = persist
        self._persist_target = persist_target
        self._persist_writer = persist_writer

    @property
    def persist(self):
        return self._persist

    @persist.setter
    def persist(self, value):
        # override only if not defined yet and given value not None
        if self._persist is None and value is not None:
            self._persist = value

    @property
    def persist_target(self):
        return self._persist_target

    @persist_target.setter
    def persist_target(self, value):
        # override only if not defined yet and given value not None
        if self._persist_target is None and value is not None:
            self._persist_target = value

    @property
    def persist_writer(self):
        return self._persist_writer

    @persist_writer.setter
    def persist_writer(self, value):
        # override only if not defined yet and given value not None
        if self._persist_writer is None and value is not None:
            self._persist_writer = value

    @instance_property
    def has_errors(self):
        """Return ``True`` if extraction error occurred on self or children
        of self, otherwise ``False``.
        """
        error = bool(self.errors)
        if not error:
            for child in self.values():
                error = child.has_errors
                if error:
                    break
        return error

    @property
    def tag(self):
        return Tag(self.translate_callable)

    def fetch(self, path):
        if isinstance(path, STR_TYPE):
            path = path.split('.')
        data = self.root
        if path[0] != data.name:
            raise KeyError('Invalid name of root element')
        __traceback_info__ = 'fetch path: {0}'.format(path)
        for key in path[1:]:
            data = data[key]
        return data

    def write(self, model, writer=None, recursiv=True):
        if self.has_errors:
            raise RuntimeError(
                'Attempt to persist data which failed to extract'
            )
        current_writer = self.persist_writer or writer
        # XXX: if persist and not current_writer
        if not current_writer:
            raise ValueError('No persistence writer found')
        if not writer:
            writer = current_writer
        if self.persist:
            target = self.persist_target
            if not target:
                target = self.name
            current_writer(model, target, self.extracted)
        if not recursiv:
            return
        for child in self.values():
            child.write(model, writer=writer, recursiv=recursiv)

    def __repr__(self):
        rep = "<RuntimeData {0}, value={1}, extracted={2}".format(
            '.'.join([str(_) for _ in self.path]),
            repr(self.value),
            repr(self.extracted)
        )
        if self.errors:
            rep += ', {0} error(s)'.format(len(self.errors))
        if len(self.attrs):
            rep += ', attrs={0}'.format(repr(self.attrs))
        rep += ' at {0}>'.format(hex(id(self))[:-1])
        return rep

    __str__ = __repr__

    @property
    def noderepr(self):
        return repr(self)


class ExtractionError(Exception):
    """Indicates problems on extraction time, such as conversion, validation
    or similar problems.
    """

    def __init__(self, msg, abort=True):
        """Initialize Exception

        :param msg: error message - usally best unicode in one-liner style.
        :param abort: If True the extraction chain continues. Default to False,
            which stops extraction.
        """
        Exception.__init__(self, msg)
        self.msg = msg
        self.abort = abort

    def __repr__(self):
        return u"ExtractionError('{0}',)".format(str(self))

    def __eq__(self, other):
        return repr(self) == repr(other)


class TBSupplementWidget(object):
    """Supplement for Tracebacks in case a widget chain part fails.

    - in conjunction with ```zExceptions``` or ```zope.exceptions```
    - goal is to ease debugging of runtime problems while processing and
      rendering.
    """

    def __init__(self, widget, func, task, descr):
        self.manageable_object = func
        try:
            name = widget.dottedpath
        except ValueError:
            name = '(name not set)'
        self.func = func
        self.name = name
        self.blueprints = widget.blueprints
        self.task = task
        self.descr = descr

    def getInfo(self, as_html=0):
        """returns additional info. zExceptions expects formatted HTML if
        as_html evaluates to True. zope.excpetions doesnt care and turns the
        result into HTML on its own.
        """
        if not as_html:
            info = '    yafowil widget processing info:\n'
            info += '    - path      : {0}\n'.format(self.name)
            info += '    - blueprints: {0}\n'.format(self.blueprints)
            info += '    - task      : {0}\n'.format(self.task)
            info += '    - descr     : {0}'.format(self.descr)
            return info
        tag = Tag(lambda x: x)
        li = tag('li', 'path: ', tag('strong', self.name))
        li += tag('li', 'blueprints: ', tag('strong', self.blueprints))
        li += tag('li', 'task: ', tag('strong', self.task))
        li += tag('li', 'description: ', tag('strong', self.descr))
        return tag('p', 'yafowil widget processing info:', tag('ul', li))


@plumbing(MappingNode, DictStorage)
class WidgetAttributes(object):

    __str__ = __repr__ = _dict__repr__

    def __init__(self, name=None, parent=None):
        self.__name__ = name
        self.__parent__ = parent

    def __getitem__(self, name):
        prefixed = '{0}.{1}'.format(self.parent.current_prefix, name)
        storage = self.storage
        if prefixed in storage:
            return storage[prefixed]
        if name in storage:
            return storage[name]
        defaults = self.parent.defaults
        if prefixed in defaults:
            return defaults[prefixed]
        if name in defaults:
            return defaults[name]
        raise KeyError((
            'Property with key "{0}" is not given on widget "{1}" (no default)'
        ).format(name, self.parent.dottedpath))


@plumbing(
    Attributes,
    MappingConstraints,
    MappingAdopt,
    Order,
    MappingNode,
    OdictStorage)
class Widget(object):
    """Base Widget Class.
    """
    attributes_factory = WidgetAttributes

    def __init__(self,
                 blueprints,
                 extractors,
                 edit_renderers,
                 display_renderers,
                 preprocessors,
                 uniquename=None,
                 value_or_getter=UNSET,
                 properties=dict(),
                 custom=dict(),
                 defaults=dict(),
                 mode='edit'):
        """Initialize the widget.

        :param blueprints: The blueprint names used to create this widget as
            list of strings.
        :param extractors: list of tuples with chain part name and callables
            extracting the data and returning it. Each extractor in chain is
            called. Expects to raise ``ExtractionException`` and provide
            error-message if something went wrong. You can call this validation.
            Need to accept some ``request`` (dict like object), ``value``
            (same as while rendering), ``uniquename`` and ``properties``, a
            dict-like. Properties gets a key ``extracted`` set, a list of
            results of previous extractors in chain-
        :param edit_renderers: List of tuples with chain part name and callable
            rendering widget in edit mode. Callable need to accept ``value``,
            ``uniquename`` and properties as dict-like. Properties gets a key
            ``rendered`` set, a list of results  of previous extractors in
            chain. Has same signature as extract.
        :param display_renderers: List of tuples with chain part name and
            callable rendering widget in display mode. Callable need to accept
            ``value``, ``uniquename`` and properties as dict-like. Properties
            gets a key ``rendered`` set, a list of results of previous
            extractors in chain. Has same signature as extract.
        :param preprocessors: List of tuples with chain part name and callable
            executed before extract or rendering. Executed only once for a
            given runtime data. Has same signature a s extract.
        :param uniquename: Id as string containing characters from a-z, A-Z,
            0-9 only. Must not start with numerical character.
        :param value_or_getter: Either a callable or the value itself. If
            callable, its called before passing to given ``renderer`` .
        :param properties: Arbitrary dict-like passed through for use in
            renderer and extractor, static data must never be modifed!
        :param custom: Definitions for custom factory chain parts.
        :param defaults: A dict with defaults value for the widgets attributes.
        :param mode: Rendering mode of widget: One out of ``edit``, ``display``,
            ``skip``.  Default is ``edit`` Expects string or callable accepting
            two parameters  ``widget`` and ``data``.
        """
        self.__name__ = uniquename
        self.__parent__ = None
        self.blueprints = blueprints
        self.getter = value_or_getter
        self.mode = mode
        self.extractors = extractors
        self.edit_renderers = edit_renderers
        self.display_renderers = display_renderers
        self.preprocessors = preprocessors or list()
        self.defaults = defaults
        self.current_prefix = ''
        # keep properties for use in dottedpath to avoid recursion errors
        self.properties = properties
        self.attrs.update(properties)
        self.custom = custom
        self._lock = RLock()

    def __call__(self, data=None, request=None):
        """Renders the widget.

        If data is passed in request is ignored! Request can't be passed in
        together with data.

        :param data: Runtime data, information collected in one run of the
            widget. If passed in, extract need to be called separate before.
            Expects either an initialized RuntimeData instance or None
            (default) to create an empty widget.
        :param request: Pass in request. if passed in it will be available at
            data. but extraction does not happen. call extract explicit before
            if needed.
        """
        if data is not None and request is not None:
            raise ValueError("if data is passed in, don't pass in request!")
        if data is None:
            if request is None:
                request = UNSET
            data = self._runpreprocessors(RuntimeData(
                name=self.name,
                request=request
            ))
        if data.mode == 'skip':
            data.rendered = u''
            return data.rendered
        elif data.mode == 'display':
            renderers = self.display_renderers
        else:
            renderers = self.edit_renderers
        if not renderers:
            raise ValueError(
                "no renderers given for widget '{0}' at mode '{1}'".format(
                    self.dottedpath,
                    data.mode
                )
            )
        with self._lock:
            try:
                for ren_name, renderer in renderers:
                    self.current_prefix = ren_name
                    __traceback_supplement__ = (
                        TBSupplementWidget,
                        self,
                        renderer,
                        'render',
                        "failed at '{0}' in mode '{1}'".format(
                            ren_name,
                            data.mode
                        )
                    )
                    data.rendered = renderer(self, data)
            finally:
                self.current_prefix = ''
        return data.rendered

    def extract(self, request, parent=None):
        """Extract the data from the request by calling the given extractors.

        :param request: Expects a dict-like object
        :param parent: Parent data
        """
        data = self._runpreprocessors(RuntimeData(
            name=self.name,
            parent=parent,
            request=request,
            persist=self.attrs.get('persist'),
            persist_target=self.attrs.get('persist_target'),
            persist_writer=self.attrs.get('persist_writer')
        ))
        # don't extract if skip mode
        if data.mode == 'skip':
            return data
        # dont't extract if display mode and no display proxy
        if data.mode == 'display':
            # display_proxy cannot be called here, currently not possible to
            # use ``attr_value``, widget not available. this causes an
            # inconsistency with the use of display_proxy inside blueprint
            # callbacks.
            # XXX: Use ``attr_value`` after signature change.
            if not self.attrs.get('display_proxy'):
                return data
        with self._lock:
            try:
                for ex_name, extractor in self.extractors:
                    self.current_prefix = ex_name
                    # update persistence settings for data node. necessary for
                    # blueprint specific factory defaults to work.
                    data.persist = self.attrs.get('persist')
                    data.persist_target = self.attrs.get('persist_target')
                    data.persist_writer = self.attrs.get('persist_writer')
                    __traceback_supplement__ = (
                        TBSupplementWidget,
                        self,
                        extractor,
                        'extract',
                        "failed at '{0}'".format(ex_name)
                    )
                    try:
                        data.extracted = extractor(self, data)
                    except ExtractionError as e:
                        data.errors.append(e)
                        if e.abort:
                            break
            finally:
                self.current_prefix = ''
        return data

    @property
    def dottedpath(self):
        if self.path[0] is None:
            raise ValueError('Root widget has no name! Pass it to factory.')
        path = list()
        node = self
        while node is not None:
            if not node.properties.get('structural'):
                path.append(node.name)
            node = node.parent
        path.reverse()
        return '.'.join(path)

    def _runpreprocessors(self, data):
        __traceback_supplement__ = (
            TBSupplementWidget,
            self,
            self._runpreprocessors,
            'run preprocessors',
            'execute'
        )
        if data.preprocessed:
            return data
        if callable(self.getter):
            data.value = self.getter(self, data)
        else:
            data.value = self.getter
        if callable(self.mode):
            data.mode = self.mode(self, data)
        else:
            data.mode = self.mode
        if data.mode not in ('edit', 'display', 'skip'):
            raise ValueError(
                "mode must be one out of 'edit', 'display', 'skip', but "
                "'{0}' was given".format(data.mode)
            )
        for pp_name, pp in self.preprocessors:
            data.current_prefix = pp_name
            __traceback_supplement__ = (
                TBSupplementWidget,
                self,
                pp,
                'preprocessor',
                "failed at '{0}'".format(pp_name)
            )
            data = pp(self, data)
        data.current_prefix = ''
        data.preprocessed = True
        return data


class Factory(object):

    def __init__(self):
        self._blueprints = dict()
        self._global_preprocessors = list()
        self._macros = dict()
        self._themes = dict()
        self.theme = 'default'
        self.defaults = dict()
        self.doc = {
            'props': dict(),
            'blueprint': dict(),
        }
        self._states = list()

    def clear(self):
        states = self._states
        self.__init__()
        self._states = states

    def push_state(self):
        self._states.append(dict(
            _blueprints=copy.deepcopy(self._blueprints),
            _global_preprocessors=copy.deepcopy(self._global_preprocessors),
            _macros=copy.deepcopy(self._macros),
            _themes=copy.deepcopy(self._themes),
            theme=copy.deepcopy(self.theme),
            defaults=copy.deepcopy(self.defaults),
            doc=copy.deepcopy(self.doc)
        ))

    def pop_state(self):
        state = self._states.pop()
        self._blueprints = state['_blueprints']
        self._global_preprocessors = state['_global_preprocessors']
        self._macros = state['_macros']
        self._themes = state['_themes']
        self.theme = state['theme']
        self.defaults = state['defaults']
        self.doc = state['doc']

    def _name_check(self, name):
        for chara in '*:#':
            if chara in name:
                raise ValueError(
                    '"{0}" as char not allowed as name.'.format(chara)
                )

    def register(
        self,
        name,
        extractors=[],
        edit_renderers=[],
        preprocessors=[],
        builders=[],
        display_renderers=[]
    ):
        """Registers a blueprint in the factory.
        """
        self._name_check(name)
        self._blueprints[name] = (
            extractors,
            edit_renderers,
            preprocessors,
            builders,
            display_renderers
        )

    def register_global_preprocessors(self, preprocessors):
        self._global_preprocessors += preprocessors

    def register_macro(self, name, blueprints, props):
        self._name_check(name)
        if isinstance(blueprints, STR_TYPE):
            blueprints = blueprints.split(':')
        self._macros[name] = blueprints, props

    def register_theme(
        self,
        themename,
        widgetname,
        resourcedir=None,
        js=[],
        css=[]
    ):
        """B/C resources registration for addon widgets.

        This function will be deprecated as of yafowil 4.0 and removed in a
        later version.

        :param themename: String or list of strings with theme names.
        :param widgetname: The widget name as string
        :param resourcedir: Registration resource directory.
        :param js: Registration list of widget related javascripts.
        :param css: Registration list of widget related stylesheets.
        """
        themenames = (
            themename
            if isinstance(themename, (list, tuple))
            else [themename]
        )
        for name in themenames:
            theme = self._themes.setdefault(name, {})
            widget_theme = theme.setdefault(widgetname, {})
            widget_theme['resourcedir'] = resourcedir
            widget_theme['js'] = js
            widget_theme['css'] = css

    def resources_for(self, widgetname, copy_resources=True):
        """B/C resources lookup.

        This function will be deprecated as of yafowil 4.0 and removed in a
        later version.

        :param widgetname: The widget name as string.
        :param copy_resources: Flag whether to return a copy of the resources.
            Defaults to True.
        """
        theme = self._themes.get(self.theme, {})
        default = self._themes.get('default', {})
        resources = theme.get(widgetname)
        if not resources:
            resources = default.get(widgetname)
        # return copy, some integrations might modify, resources are static
        if copy_resources:
            return copy.deepcopy(resources)
        return resources

    def register_resources(self, theme_name, widget_name, resources):
        """Register resource group containing theme related scripts and styles
        for widget.

        :param theme_name: String or list of strings with theme names.
        :param widget_name: The widget name as string
        :param resources: webresource.ResourceGroup containing widget resources.
        """
        theme_names = (
            theme_name
            if isinstance(theme_name, (list, tuple))
            else [theme_name]
        )
        for name in theme_names:
            theme = self._themes.setdefault(name, {})
            widget_theme = theme.setdefault(widget_name, {})
            widget_theme['resources'] = resources

    def get_resources(self, widget_name=None, copy_resources=True, exclude=[]):
        """Resources lookup.

        :param widget_name: The widget name. If None, all registered resources
            are returned.
        :param copy_resources: Flag whether to return a copy of the resources.
            Defaults to True.
        :param exclude: List of widget names to exclude. Only takes effect if
            ``widget_name`` not given.
        """
        if widget_name:
            resources = self._get_widget_resources(widget_name)
        else:
            resources = self._get_all_resources(exclude)
        if copy_resources:
            return resources.copy()
        return resources

    def _get_all_resources(self, exclude):
        widget_names = set()
        for theme in self._themes.values():
            widget_names.update(theme)
        resources = wr.ResourceGroup(name='yafowil')
        for widget_name in sorted(widget_names):
            if widget_name in exclude:
                continue
            resources.add(self._get_widget_resources(widget_name))
        return resources

    def _get_widget_resources(self, widget_name):
        theme = self._themes.get(self.theme, {})
        default = self._themes.get('default', {})
        widget_theme = theme.get(widget_name, default.get(widget_name, {}))
        return widget_theme.get('resources', wr.ResourceGroup(name=widget_name))

    def _expand_blueprints(self, blueprints, props):
        result = list()
        if isinstance(blueprints, STR_TYPE):
            blueprints = blueprints.split(':')
        for blueprint in blueprints:
            if blueprint.startswith('#'):
                macro_name = blueprint[1:]
                if macro_name not in self._macros:
                    raise ValueError(
                        "Macro named '{}' is not registered in factory".format(
                            macro_name
                        )
                    )
                macro_chain, macro_props = self._macros[macro_name]
                for key in macro_props:
                    if key not in props:
                        props[key] = macro_props[key]
                expanded, props = self._expand_blueprints(macro_chain, props)
                result += expanded
            else:
                result.append(blueprint)
        return result, props

    def __call__(
        self,
        blueprints,
        name=None,
        value=UNSET,
        props=dict(),
        custom=dict(),
        mode='edit'
    ):
        """Creates a widget from blueprints.

        ``blueprints``
            chain of blueprint names. Either a colon separated string or a list
            of strings defining which widget to build.
            I.e. creating a simple text widget use ``"text"``. Wrapping a text
            blueprint with a label use ``"label:text"`` or
            ``["label", "text"]``.
            Custom blueprints not registered in the registry are added using
            asterisk syntax. I.e. inject a validating extractor with
            ``label:*myextractor:text`` and use the ``custom`` keyword argument
            (see below).

        ``name``
            a name for the widget. optional. if not set it has to be set later
            at the  ``__name__`` attribute of the widget.
            if not set the widget wont work.

        ``value``
            a value or a callable used as getter. If callable is given it has
            to accept widget and runtime-data as argument.

        ``props``
            dict-like object containing properties to be copied to widgets
            attributes.

        ``custom``
            dict, where keys are matching to asterisk prefixed custom chains.
            each chains part is tuple with 4 lists of callables: extractors,
            edit_renderers, preprocessors, builders, display_renderers.

        ``mode``
            either a callable (widget and runtime-data as argument) or string
            returning one of 'edit', 'display', 'skip' or direct value.
            Defaults to 'edit'.
        """
        extractors = list()
        edit_renderers = list()
        disp_renderers = list()
        preprocessors = list()
        builders = list()
        blueprints, props = self._expand_blueprints(blueprints, props)
        for blueprint in blueprints:
            if blueprint.startswith('*'):
                part_name = blueprint[1:]
                if type(custom[part_name]) in ITER_TYPES:
                    if len(custom[part_name]) < 5:
                        # BBB:
                        ex, eren, pre, bui = custom[part_name]
                        dren = []
                    else:
                        ex, eren, pre, bui, dren = custom[part_name]
                else:  # expect dict
                    ex = custom[part_name].get('extractors', list())
                    eren = custom[part_name].get('edit_renderers', list())
                    pre = custom[part_name].get('preprocessors', list())
                    bui = custom[part_name].get('builders', list())
                    dren = custom[part_name].get('display_renderers', list())
            else:
                part_name = blueprint
                ex, eren, pre, bui, dren = self._blueprints[part_name]
            extractors = [(part_name, _) for _ in ex] + extractors
            edit_renderers = [(part_name, _) for _ in eren] + edit_renderers
            disp_renderers = [(part_name, _) for _ in dren] + disp_renderers
            preprocessors = preprocessors + [(part_name, _) for _ in pre]
            builders = builders + [(part_name, _) for _ in bui]
        global_pre = [('__GLOBAL__', _) for _ in self._global_preprocessors]
        widget = Widget(
            blueprints,
            extractors,
            edit_renderers,
            disp_renderers,
            global_pre + preprocessors,
            uniquename=name,
            value_or_getter=value,
            properties=props,
            custom=custom,
            defaults=self.defaults,
            mode=mode
        )
        for part_name, builder_func in builders:
            widget.current_prefix = part_name
            builder_func(widget, self)
            widget.current_prefix = ''
        return widget

    def extractors(self, name):
        return self._blueprints[name][0]

    def renderers(self, name):
        raise RuntimeError(
            'Deprecated since 1.2, use edit_renderers or display_renderers'
        )

    def edit_renderers(self, name):
        return self._blueprints[name][1]

    def display_renderers(self, name):
        return self._blueprints[name][4]

    def preprocessors(self, name):
        return self._global_preprocessors + self._blueprints[name][2]

    def builders(self, name):
        return self._blueprints[name][3]


factory = Factory()


def fetch_value(widget, data):
    """Fetch extracted, either form-data, value or default .
    """
    if data.extracted is not UNSET:
        return data.extracted
    if data.value is not UNSET:
        return data.value
    return attr_value('default', widget, data)



import os


class DummyTranslationStringFactory(object):
    """Dummy Translations string factory.
    """

    def __init__(self, domain):
        self.domain = domain

    def __call__(self, message, default='', mapping={}):
        """Directly create message and return it as is.
        """
        message = default or message
        if mapping:
            for k, v in mapping.items():
                message = message.replace('${' + k + '}', str(v))
        return message


# try to import framework corresponding translation string factories if no
# test run
if not os.environ.get('TESTRUN_MARKER'):
    # pyramid related
    try:
        from pyramid.i18n import TranslationStringFactory as TSF
    except ImportError:
        # zope related
        try:
            from zope.i18nmessageid import MessageFactory as TSF
        # fallback ti dummy
        except ImportError:
            TSF = DummyTranslationStringFactory
# test run, use dummy translation string factory
else:
    TSF = DummyTranslationStringFactory  # pragma: no cover


_ = TSF('yafowil')



from node.utils import UNSET
from odict import odict
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.tests import YafowilTestCase
from yafowil.tests import fxml
import yafowil.common
import yafowil.compound
import yafowil.table  # noqa


class TestTable(YafowilTestCase):
    # Table elements are available as blueprints, often one wanta to
    # organize form elements inside a table, providing pretty looking forms

    def test_table_blueprint(self):
        table = factory(
            'table',
            name='foo')
        self.assertEqual(table(), '<table></table>')

        table = factory(
            'table',
            name='foo_table',
            props={
                'id': 'id',
                'class': 'css'
            })
        self.assertEqual(table(), '<table class="css" id="id"></table>')

        table = factory(
            'table',
            name='foo',
            mode='display')
        self.assertEqual(table(), '<table></table>')

    def test_thead_blueprint(self):
        thead = factory(
            'thead',
            name='foo')
        self.assertEqual(thead(), '<thead></thead>')

        thead = factory(
            'thead',
            name='foo',
            mode='display')
        self.assertEqual(thead(), '<thead></thead>')

    def test_tbody_blueprint(self):
        tbody = factory(
            'tbody',
            name='foo')
        self.assertEqual(tbody(), '<tbody></tbody>')

        tbody = factory(
            'tbody',
            name='foo',
            mode='display')
        self.assertEqual(tbody(), '<tbody></tbody>')

    def test_tr_blueprint(self):
        tr = factory(
            'tr',
            name='foo')
        self.assertEqual(tr(), '<tr></tr>')

        tr = factory(
            'tr',
            name='foo',
            mode='display')
        self.assertEqual(tr(), '<tr></tr>')

        tr = factory(
            'tr',
            name='foo',
            props={
                'id': 'id',
                'class': 'css'
            })
        self.assertEqual(tr(), '<tr class="css" id="id"></tr>')

    def test_th_blueprint(self):
        th = factory(
            'th',
            name='foo')
        self.assertEqual(th(), '<th></th>')

        th = factory(
            'th',
            name='foo',
            mode='display')
        self.assertEqual(th(), '<th></th>')

        th = factory(
            'th',
            name='foo',
            props={
                'id': 'id',
                'class': 'css',
                'colspan': 2,
                'rowspan': 2,
            })
        self.assertEqual(
            th(),
            '<th class="css" colspan="2" id="id" rowspan="2"></th>'
        )

    def test_td_blueprint(self):
        td = factory(
            'td',
            name='foo')
        self.assertEqual(td(), '<td></td>')

        td = factory(
            'td',
            name='foo',
            mode='display')
        self.assertEqual(td(), '<td></td>')

        td = factory(
            'td',
            name='foo',
            props={
                'id': 'id',
                'class': 'css',
                'colspan': 2,
                'rowspan': 2,
            })
        self.assertEqual(
            td(),
            '<td class="css" colspan="2" id="id" rowspan="2"></td>'
        )

    def test_render_table(self):
        form = factory(
            'form',
            name='myform',
            props={
                'action': 'myaction',
            })
        form['table'] = factory('table')
        form['table']['row1'] = factory('tr')
        form['table']['row1']['field1'] = factory(
            'td:text',
            name='field1')
        self.checkOutput("""
        <form action="myaction" enctype="multipart/form-data" id="form-myform"
              method="post" novalidate="novalidate">
          <table>
            <tr>
              <td>
                <input class="text" id="input-myform-table-row1-field1"
                       name="myform.table.row1.field1" type="text" value=""/>
              </td>
            </tr>
          </table>
        </form>
        """, fxml(form()))

    def test_table_with_structural(self):
        # Build same table again but set some nodes structural. This is
        # considered in ``Widget.dottedpath``
        form = factory(
            'form',
            name='mytableform',
            props={
                'action': 'mytableaction',
            })
        form['table'] = factory(
            'table',
            props={
                'structural': True
            })
        form['table']['row1'] = factory(
            'tr',
            props={
                'structural': True
            })
        # note: td is used in a blueprint chain here
        form['table']['row1']['field1'] = factory(
            'td:error:text',
            props={
                'required': 'Field 1 is required',
            }
        )
        self.checkOutput("""
        <form action="mytableaction" enctype="multipart/form-data"
              id="form-mytableform" method="post" novalidate="novalidate">
          <table>
            <tr>
              <td>
                <input class="required text" id="input-mytableform-field1"
                       name="mytableform.field1" required="required"
                       type="text" value=""/>
              </td>
            </tr>
          </table>
        </form>
        """, fxml(form()))

        data = form.extract({})
        self.assertEqual(data.name, 'mytableform')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, odict([('field1', UNSET)]))
        self.assertEqual(data.errors, [])

        field_data = data['field1']
        self.assertEqual(field_data.name, 'field1')
        self.assertEqual(field_data.value, UNSET)
        self.assertEqual(field_data.extracted, UNSET)
        self.assertEqual(field_data.errors, [])

        data = form.extract({'mytableform.field1': ''})
        self.assertEqual(data.name, 'mytableform')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, odict([('field1', '')]))
        self.assertEqual(data.errors, [])
        self.assertTrue(data.has_errors)

        field_data = data['field1']
        self.assertEqual(field_data.name, 'field1')
        self.assertEqual(field_data.value, UNSET)
        self.assertEqual(field_data.extracted, '')
        self.assertEqual(field_data.errors, [
            ExtractionError('Field 1 is required')
        ])

        self.checkOutput("""
        <form action="mytableaction" enctype="multipart/form-data"
              id="form-mytableform" method="post" novalidate="novalidate">
          <table>
            <tr>
              <td>
                <div class="error">
                  <div class="errormessage">Field 1 is required</div>
                  <input class="required text" id="input-mytableform-field1"
                         name="mytableform.field1" required="required"
                         type="text" value=""/>
                </div>
              </td>
            </tr>
          </table>
        </form>
        """, fxml(form(data)))

    def test_table_with_compound_td(self):
        # Create table with 'td' as compound
        form = factory(
            'form',
            name='mytableform',
            props={
                'action': 'mytableaction',
            })
        form['table'] = factory(
            'table',
            props={
                'structural': True
            })
        form['table']['row1'] = factory(
            'tr',
            props={
                'structural': True
            })
        form['table']['row1']['td1'] = factory(
            'td',
            props={
                'structural': True
            })
        form['table']['row1']['td1']['field1'] = factory(
            'error:text',
            props={
                'required': 'Field 1 is required',
            }
        )
        self.checkOutput("""
        <form action="mytableaction" enctype="multipart/form-data"
              id="form-mytableform" method="post" novalidate="novalidate">
          <table>
            <tr>
              <td>
                <input class="required text" id="input-mytableform-field1"
                       name="mytableform.field1" required="required"
                       type="text" value=""/>
              </td>
            </tr>
          </table>
        </form>
        """, fxml(form()))

        data = form.extract({})
        self.assertEqual(data.name, 'mytableform')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, odict([('field1', UNSET)]))
        self.assertEqual(data.errors, [])

        field_data = data['field1']
        self.assertEqual(field_data.name, 'field1')
        self.assertEqual(field_data.value, UNSET)
        self.assertEqual(field_data.extracted, UNSET)
        self.assertEqual(field_data.errors, [])

        data = form.extract({'mytableform.field1': ''})
        self.assertEqual(data.name, 'mytableform')
        self.assertEqual(data.value, UNSET)
        self.assertEqual(data.extracted, odict([('field1', '')]))
        self.assertEqual(data.errors, [])
        self.assertTrue(data.has_errors)

        field_data = data['field1']
        self.assertEqual(field_data.name, 'field1')
        self.assertEqual(field_data.value, UNSET)
        self.assertEqual(field_data.extracted, '')
        self.assertEqual(field_data.errors, [
            ExtractionError('Field 1 is required')
        ])

        self.checkOutput("""
        <form action="mytableaction" enctype="multipart/form-data"
              id="form-mytableform" method="post" novalidate="novalidate">
          <table>
            <tr>
              <td>
                <div class="error">
                  <div class="errormessage">Field 1 is required</div>
                  <input class="required text" id="input-mytableform-field1"
                         name="mytableform.field1" required="required"
                         type="text" value=""/>
                </div>
              </td>
            </tr>
          </table>
        </form>
        """, fxml(form(data)))



# -*- coding: utf-8 -*-
from node.utils import UNSET
from yafowil.base import ExtractionError
from yafowil.base import factory
from yafowil.base import fetch_value
from yafowil.compat import BYTES_TYPE
from yafowil.compat import IS_PY2
from yafowil.compat import ITER_TYPES
from yafowil.compat import LONG_TYPE
from yafowil.compat import STR_TYPE
from yafowil.compat import UNICODE_TYPE
from yafowil.tsf import _
from yafowil.utils import EMPTY_VALUE
from yafowil.utils import as_data_attrs
from yafowil.utils import attr_value
from yafowil.utils import convert_value_to_datatype
from yafowil.utils import convert_values_to_datatype
from yafowil.utils import css_managed_props
from yafowil.utils import cssclasses
from yafowil.utils import cssid
from yafowil.utils import managedprops
from yafowil.utils import vocabulary
import re
import uuid


###############################################################################
# common defaults
###############################################################################

factory.defaults['default'] = UNSET
factory.doc['props']['default'] = """\
Default value for rendering.
"""

factory.defaults['class'] = None
factory.doc['props']['class'] = """\
Common CSS-class to put on.
"""

factory.defaults['class_add'] = None
factory.doc['props']['class_add'] = """\
Additional CSS-class to put on.
"""

factory.defaults['error_class'] = None
factory.doc['props']['error_class'] = """\
CSS-class to put on in case of error after extraction.
"""

factory.defaults['error_class_default'] = 'error'
factory.doc['props']['error_class_default'] = """\
Fallback CSS-class to put on in case of error if no specific class was
given.
"""

factory.defaults['valid_class'] = None
factory.doc['props']['valid_class'] = """\
CSS-class to put on in case of valid value after extraction.
"""

factory.defaults['valid_class_default'] = 'valid'
factory.doc['props']['valid_class_default'] = """\
Fallback CSS-class to put on in case of valid value if no specific class was
given.
"""

factory.defaults['autofocus'] = None
factory.doc['props']['autofocus'] = """\
Whether this field gets the focus automatically or not (if browser supports
it).
"""

factory.defaults['autocomplete'] = None
factory.doc['props']['autocomplete'] = """\
Switch autocomplete explizit to ``on`` or ``off``.
"""

factory.defaults['placeholder'] = None
factory.doc['props']['placeholder'] = """\
Whether this input has a placeholder value or not (if browser supports it).
"""

factory.doc['props']['emptyvalue'] = """\
If configured and received value in request is empty, return as extracted
value.
"""

factory.defaults['datatype'] = None
factory.doc['props']['datatype'] = """\
Callable for converting extracted value to output datatype.

``datatype`` can also be defined as string with value out of ``'str'``,
``'unicode'``, ``'int'``, ``'integer'``, ``'long'``, ``'float'`` or
``'uuid'``.

ATTENTION: In python 3, string identifiers should not be used at all, but if,
    the following must be known:
        * ``'str'`` refers to ``bytes`` type.
        * ``'unicode'`` refers to ``str`` type.
        * ``'long'`` refers to ``int`` type.

Custom converter callables must raise one out of the following exceptions if
conversion fails:
    * ``ValueError``
    * ``UnicodeDecodeError``
    * ``UnicodeEncodeError``
"""

factory.defaults['allowed_datatypes'] = UNSET
factory.doc['props']['allowed_datatypes'] = """\
List of allowed datatypes. If ``UNSET``, datatype converters are not
restricted.
"""

factory.defaults['datatype_message'] = None
factory.doc['props']['datatype_message'] = """\
Custom extraction error message if ``datatype`` conversion fails.
"""

factory.defaults['required'] = False
factory.doc['props']['required'] = """\
Whether this value is required or not.
"""

factory.defaults['required_message'] = _(
    'required_message', default=u'Mandatory field was empty')
factory.doc['props']['required_message'] = """\
Message to be shown if required condition was not met.
"""

factory.defaults['required_class'] = None
factory.doc['props']['required_class'] = """\
CSS-class to put on in case if required condition was not met.
"""

factory.defaults['type'] = None
factory.doc['props']['type'] = """\
HTML type attribute.
"""

factory.defaults['size'] = None
factory.doc['props']['size'] = """\
Allowed input size.
"""

factory.defaults['maxlength'] = None
factory.doc['props']['maxlength'] = """\
Input maxlength.
"""

factory.defaults['disabled'] = None
factory.doc['props']['disabled'] = """\
Disables input.
"""

factory.defaults['required_class_default'] = 'required'
factory.doc['props']['required_class_default'] = """\
CSS-class to apply if required condition was not met - if no specific class
was given.
"""

factory.defaults['template'] = '%s'
factory.doc['props']['template'] = """\
Format string with pythons built-in string format template. If a callable
is given it will be used instead and is called with ``widget`` and ``data`` as
parameters.
"""

factory.defaults['title'] = None
factory.doc['props']['title'] = """\
Optional help text to be rendered in the title attribute.
"""

factory.defaults['data'] = dict()
factory.doc['props']['data'] = """\
Additional data rendered as HTML5 data attributes on DOM Element.
"""

factory.defaults['display_proxy'] = False
factory.doc['props']['display_proxy'] = """\
If 'True' and widget mode 'display', widget value gets rendered as hidden
input.
"""


###############################################################################
# generic
###############################################################################

def generic_extractor(widget, data):
    """Extract raw value from request by ``widget.dottedpath``.

    Return ``UNSET`` if widget not present in request.
    """
    __managed_props = []  # noqa
    try:
        return data.request[widget.dottedpath]
    except KeyError:
        return UNSET


@managedprops('required', 'required_message')
def generic_required_extractor(widget, data):
    """Validate required.

    If ``required`` is set and some raw value has been extracted,
    evaluate ``data.extracted`` to boolean. Raise ``ExtractionError`` if
    ``False``.
    """
    required = attr_value('required', widget, data)
    if not required or bool(data.extracted) or data.extracted is UNSET:
        return data.extracted
    if isinstance(required, STR_TYPE):
        raise ExtractionError(required)
    raise ExtractionError(attr_value('required_message', widget, data))


@managedprops('emptyvalue')
def generic_emptyvalue_extractor(widget, data):
    """Return emptyvalue if widget present in request and raw value is empty.
    """
    try:
        if not data.request[widget.dottedpath]:
            return attr_value('emptyvalue', widget, data, data.extracted)
    except KeyError:
        pass
    return data.extracted


DATATYPE_LABELS = {
    BYTES_TYPE: _('datatype_str', default='string'),
    UNICODE_TYPE: _('datatype_unicode', default='unicode'),
    int: _('datatype_integer', default='integer'),
    float: _('datatype_float', default='floating point number'),
    uuid.UUID: _('datatype_uuid', default='UUID')
}
if IS_PY2:
    DATATYPE_LABELS[LONG_TYPE] = _('datatype_long', default='long integer')  # pragma: no cover
else:
    DATATYPE_LABELS[LONG_TYPE] = DATATYPE_LABELS[int]  # pragma: no cover
# B/C
DATATYPE_LABELS['str'] = DATATYPE_LABELS[BYTES_TYPE]
DATATYPE_LABELS['unicode'] = DATATYPE_LABELS[UNICODE_TYPE]
DATATYPE_LABELS['int'] = DATATYPE_LABELS[int]
DATATYPE_LABELS['integer'] = DATATYPE_LABELS[int]
DATATYPE_LABELS['long'] = DATATYPE_LABELS[LONG_TYPE]
DATATYPE_LABELS['float'] = DATATYPE_LABELS[float]
DATATYPE_LABELS['uuid'] = DATATYPE_LABELS[uuid.UUID]


@managedprops('datatype', 'allowed_datatypes',
              'datatype_message', 'emptyvalue')
def generic_datatype_extractor(widget, data):
    """Convert extracted value to ``datatype``.

    If extracted value is ``UNSET`` return ``UNSET``.
    If no ``datatype`` given, return extracted value.
    Otherwise try to convert value to given ``datatype`` and return the
    converted value or raise an ``ExtractionError`` if conversion fails.
    Extraction error message can be customized with ``datatype_message``
    property. Value can also be a list, then all items inside the list are
    converted.
    """
    extracted = data.extracted
    if extracted is UNSET:
        return extracted
    # datatype is one of the rare cases where an attribute callable not follows
    # the typical signature taking widget and data as arguments, but is just
    # called with the value.
    datatype = widget.attrs.get('datatype', None)
    if not datatype:
        return extracted
    allowed_datatypes = attr_value('allowed_datatypes', widget, data)
    if allowed_datatypes and datatype not in allowed_datatypes:
        raise ValueError('Datatype not allowed: "{0}"'.format(datatype))
    try:
        emptyvalue = attr_value('emptyvalue', widget, data, EMPTY_VALUE)
        return convert_values_to_datatype(
            extracted,
            datatype,
            empty_value=emptyvalue
        )
    except KeyError:
        raise ValueError('Datatype unknown: "{0}"'.format(datatype))
    except (ValueError, UnicodeEncodeError, UnicodeDecodeError):
        datatype_message = attr_value('datatype_message', widget, data)
        if not datatype_message:
            datatype_label = DATATYPE_LABELS.get(datatype)
            if not datatype_label:
                datatype_message = _(
                    'generic_datatype_message',
                    default=u'Input conversion failed.'
                )
            else:
                if data.tag.translate:
                    datatype_label = data.tag.translate(datatype_label)
                datatype_message = _(
                    'standard_datatype_message',
                    default=u'Input is not a valid ${datatype}.',
                    mapping={
                        'datatype': datatype_label
                    }
                )
        raise ExtractionError(datatype_message)


def input_attributes_common(widget, data, excludes=list(), value=None):
    if value is None:
        value = fetch_value(widget, data)
    if isinstance(value, STR_TYPE):
        value = value.replace('"', '&quot;')
    autofocus = attr_value('autofocus', widget, data) and 'autofocus' or None
    disabled = attr_value('disabled', widget, data)
    disabled = bool(disabled) and 'disabled' or None
    required = attr_value('required', widget, data) and 'required' or None
    input_attrs = {
        'autofocus': autofocus,
        'class_': cssclasses(widget, data),
        'disabled': disabled,
        'id': cssid(widget, 'input'),
        'name_': widget.dottedpath,
        'placeholder': attr_value('placeholder', widget, data),
        'required': required,
        'size': attr_value('size', widget, data),
        'maxlength': attr_value('maxlength', widget, data),
        'title': attr_value('title', widget, data),
        'type': attr_value('type', widget, data),
        'value': value,
    }
    input_attrs.update(as_data_attrs(attr_value('data', widget, data)))
    for attr_name in excludes:
        del input_attrs[attr_name]
    return input_attrs


def input_attributes_full(widget, data, value=None):
    input_attrs = input_attributes_common(widget, data, value=value)
    input_attrs['autocomplete'] = attr_value(
        'autocomplete', widget, data)
    if attr_value('type', widget, data) in ['range', 'number']:
        input_attrs['min'] = attr_value('min', widget, data)
        input_attrs['max'] = attr_value('max', widget, data)
        input_attrs['step'] = attr_value('step', widget, data)
    return input_attrs


@managedprops(*css_managed_props)
def input_generic_renderer(widget, data, custom_attrs={}):
    """Generic HTML ``input`` tag render.
    """
    input_attrs = input_attributes_full(widget, data)
    input_attrs.update(custom_attrs)
    return data.tag('input', **input_attrs)


# multivalued is not documented, because its only valid for specific blueprints
@managedprops('display_proxy')
def display_proxy_renderer(widget, data):
    rendered = data.rendered
    if attr_value('display_proxy', widget, data):
        orgin_type = attr_value('type', widget, data)
        widget.attrs['type'] = 'hidden'
        value = fetch_value(widget, data)
        multivalued = attr_value('multivalued', widget, data)
        if multivalued and isinstance(value, STR_TYPE):
            value = [value]
        if multivalued or type(value) in ITER_TYPES:
            for val in value:
                input_attrs = input_attributes_full(widget, data, value=val)
                rendered += data.tag('input', **input_attrs)
        else:
            rendered += input_generic_renderer(widget, data)
        if orgin_type:
            widget.attrs['type'] = orgin_type
        else:
            del widget.attrs['type']
    return rendered


@managedprops('template', 'class')
def generic_display_renderer(widget, data, value=None):
    """Generic display renderer to render a value.
    """
    if callable(widget.attrs['template']):
        content = widget.attrs['template'](widget, data)
    elif value is None:
        value = fetch_value(widget, data)
        if value in [UNSET, None]:
            value = u''
        content = widget.attrs['template'] % value
    else:
        content = widget.attrs['template'] % value
    attrs = {
        'id': cssid(widget, 'display'),
        'class_': 'display-{0}'.format(
            attr_value('class', widget, data) or 'generic'
        )
    }
    attrs.update(as_data_attrs(attr_value('data', widget, data)))
    return data.tag('div', content, **attrs)


def empty_display_renderer(widget, data):
    """Display renderer which renders an empty string.
    """
    return data.rendered or u''


def generic_positional_rendering_helper(tagname, message, attrs, rendered, pos,
                                        tag):
    """returns new tag with rendered content dependent on position

    tagname
        name of new tag, ie. div, p, label and so on

    message
        text included in new tag

    attrs
        attributes on new tag

    rendered
        prior rendered text

    pos
        position how to place the newtag relative to the prior rendered:
        'before'='<newtag>message</newtag>rendered',
        'after' ='<newtag>message</newtag>'
        'inner-before'= <newtag>message rendered</newtag>
        'inner-after'= <newtag>rendered message</newtag>
    """
    if pos not in ['before', 'after', 'inner-before', 'inner-after']:
        raise ValueError('Invalid value for position "{0}"'.format(pos))
    if pos.startswith('inner'):
        if pos.endswith('before'):
            inner = message, rendered
        else:
            inner = rendered, message
        return tag(tagname, *inner, **attrs)
    else:
        newtag = tag(tagname, message, **attrs)
        if pos == 'before':
            return newtag + rendered
        return rendered + newtag


###############################################################################
# tag
###############################################################################

@managedprops('tag', 'text', 'data', *css_managed_props)
def tag_renderer(widget, data):
    """Renderer for HTML tags.
    """
    attrs = {
        'id': cssid(widget, 'tag'),
        'class_': cssclasses(widget, data),
    }
    attrs.update(as_data_attrs(attr_value('data', widget, data)))
    tag = attr_value('tag', widget, data)
    text = attr_value('text', widget, data)
    return data.tag(tag, text, **attrs)


factory.register(
    'tag',
    edit_renderers=[tag_renderer],
    display_renderers=[tag_renderer])

factory.doc['blueprint']['tag'] = """\
Render HTML tags with text. Useful for rendering headings etc.
"""

factory.doc['props']['tag.tag'] = """\
HTML tag name.
"""

factory.doc['props']['tag.text'] = """\
Tag contents.
"""


###############################################################################
# text
###############################################################################

@managedprops('data', 'title', 'size', 'disabled', 'autofocus',
              'placeholder', 'autocomplete', *css_managed_props)
def text_edit_renderer(widget, data):
    return input_generic_renderer(widget, data)


factory.register(
    'text',
    extractors=[
        generic_extractor,
        generic_required_extractor,
        generic_emptyvalue_extractor,
        generic_datatype_extractor,
    ],
    edit_renderers=[text_edit_renderer],
    display_renderers=[
        generic_display_renderer,
        display_proxy_renderer
    ])

factory.doc['blueprint']['text'] = """\
One line text input blueprint.
"""

factory.defaults['text.type'] = 'text'
factory.doc['props']['text.type'] = """\
Type of input tag.
"""

factory.defaults['text.required_class'] = 'required'

factory.defaults['text.default'] = ''

factory.defaults['text.class'] = 'text'

factory.defaults['text.disabled'] = False
factory.doc['props']['text.disabled'] = """\
Flag  input field is disabled.
"""

factory.defaults['text.persist'] = True


###############################################################################
# hidden
###############################################################################

factory.register(
    'hidden',
    extractors=[
        generic_extractor,
        generic_emptyvalue_extractor,
        generic_datatype_extractor
    ],
    edit_renderers=[input_generic_renderer],
    display_renderers=[empty_display_renderer])

factory.doc['blueprint']['hidden'] = """\
Hidden input blueprint.
"""

factory.defaults['hidden.type'] = 'hidden'
factory.doc['props']['hidden.type'] = """\
Type of input tag.
"""

factory.defaults['hidden.default'] = ''

factory.defaults['hidden.class'] = 'hidden'

factory.defaults['hidden.persist'] = True


###############################################################################
# proxy
###############################################################################

@managedprops(*css_managed_props)
def input_proxy_renderer(widget, data):
    """Render hidden input ignoring ``widget.dottedpath``, just using widget
    name.
    """
    tag = data.tag
    value = data.value
    if data.request is not UNSET and data.request.get(widget.__name__):
        value = data.request.get(widget.__name__)
    input_attrs = {
        'type': 'hidden',
        'value': value,
        'name_': widget.__name__,
        'id': cssid(widget, 'input'),
        'class_': cssclasses(widget, data),
    }
    return tag('input', **input_attrs)


factory.register(
    'proxy',
    extractors=[
        generic_extractor,
        generic_emptyvalue_extractor,
        generic_datatype_extractor
    ],
    edit_renderers=[input_proxy_renderer],
    display_renderers=[empty_display_renderer])

factory.doc['blueprint']['proxy'] = """\
Bypass arguments out of form namespace using a hidden field.
"""

factory.defaults['proxy.class'] = None


###############################################################################
# textarea
###############################################################################

def textarea_attributes(widget, data):
    autofocus = attr_value('autofocus', widget, data) and 'autofocus' or None
    disabled = attr_value('disabled', widget, data) and 'disabled' or None
    readonly = attr_value('readonly', widget, data) and 'readonly' or None
    required = attr_value('required', widget, data) and 'required' or None
    ta_attrs = {
        'autofocus': autofocus,
        'class_': cssclasses(widget, data),
        'cols': attr_value('cols', widget, data),
        'disabled': disabled,
        'id': cssid(widget, 'input'),
        'title': attr_value('title', widget, data),
        'name_': widget.dottedpath,
        'placeholder': attr_value('placeholder', widget, data),
        'readonly': readonly,
        'required': required,
        'rows': attr_value('rows', widget, data),
        'wrap': attr_value('wrap', widget, data),
    }
    ta_attrs.update(as_data_attrs(attr_value('data', widget, data)))
    return ta_attrs


textarea_managed_props = [
    'data', 'title', 'autofocus', 'cols', 'disabled',
    'placeholder', 'readonly', 'required', 'rows', 'wrap',
] + css_managed_props


@managedprops(*textarea_managed_props)
def textarea_renderer(widget, data, custom_attrs={}):
    """Renders text area.
    """
    tag = data.tag
    area_attrs = textarea_attributes(widget, data)
    area_attrs.update(custom_attrs)
    value = fetch_value(widget, data)
    if value is None:
        value = ''
    return tag('textarea', value, **area_attrs)


factory.register(
    'textarea',
    extractors=[
        generic_extractor,
        generic_required_extractor,
        generic_emptyvalue_extractor
    ],
    edit_renderers=[textarea_renderer],
    display_renderers=[
        generic_display_renderer,
        display_proxy_renderer
    ])

factory.doc['blueprint']['textarea'] = """\
HTML textarea blueprint.
"""

factory.defaults['textarea.default'] = ''

factory.defaults['textarea.wrap'] = None
factory.doc['props']['textarea.wrap'] = """\
Either ``soft``, ``hard``, ``virtual``, ``physical`` or  ``off``.
"""

factory.defaults['textarea.class'] = 'textarea'

factory.defaults['textarea.cols'] = 80
factory.doc['props']['textarea.cols'] = """\
Number of characters.
"""

factory.defaults['textarea.rows'] = 25
factory.doc['props']['textarea.rows'] = """\
Number of lines.
"""

factory.defaults['textarea.readonly'] = None
factory.doc['props']['textarea.readonly'] = """\
Flag textarea is readonly.
"""

factory.defaults['textarea.persist'] = True


###############################################################################
# lines
###############################################################################

def lines_extractor(widget, data):
    """Extract textarea value as list of lines.
    """
    extracted = data.extracted
    if not extracted:
        return list()
    return extracted.split('\n')


@managedprops(*textarea_managed_props)
def lines_edit_renderer(widget, data):
    """Renders text area with list value as lines.
    """
    tag = data.tag
    area_attrs = textarea_attributes(widget, data)
    value = fetch_value(widget, data)
    if value is None:
        value = u''
    else:
        value = u'\n'.join(value)
    return tag('textarea', value, **area_attrs)


@managedprops('class', 'data')
def lines_display_renderer(widget, data):
    value = fetch_value(widget, data)
    if type(value) in ITER_TYPES and not value:
        value = u''
    attrs = {
        'id': cssid(widget, 'display'),
        'class_': 'display-{0}'.format(attr_value('class', widget, data))
    }
    attrs.update(as_data_attrs(attr_value('data', widget, data)))
    content = u''
    for line in value:
        content += data.tag('li', line)
    return data.tag('ul', content, **attrs)


factory.register(
    'lines',
    extractors=[
        generic_extractor,
        generic_required_extractor,
        lines_extractor,
        generic_emptyvalue_extractor,
        generic_datatype_extractor,
    ],
    edit_renderers=[lines_edit_renderer],
    display_renderers=[
        lines_display_renderer,
        display_proxy_renderer
    ])

factory.doc['blueprint']['lines'] = """\
Lines blueprint. Renders a textarea and extracts lines as list.
"""

factory.defaults['lines.default'] = ''

factory.defaults['lines.class'] = 'lines'

factory.defaults['lines.wrap'] = None
factory.doc['props']['lines.wrap'] = """\
Either ``soft``, ``hard``, ``virtual``, ``physical`` or  ``off``.
"""

factory.defaults['lines.cols'] = 40
factory.doc['props']['lines.cols'] = """\
Number of characters.
"""

factory.defaults['lines.rows'] = 8
factory.doc['props']['lines.rows'] = """\
Number of lines.
"""

factory.defaults['lines.readonly'] = None
factory.doc['props']['lines.readonly'] = """\
Flag textarea is readonly.
"""

factory.defaults['lines.persist'] = True


###############################################################################
# password
###############################################################################

@managedprops('minlength')
def minlength_extractor(widget, data):
    """Validate minlength of a string input.

    Only perform if ``minlength`` property is set.

    Properties:

    ``minlength``
        Minimum length of string as int.
    """
    val = data.extracted
    if val is UNSET:
        return val
    minlength = attr_value('minlength', widget, data, -1)
    if minlength != -1:
        if len(val) < minlength:
            message = _('minlength_extraction_error',
                        default=u'Input must have at least ${len} characters.',
                        mapping={'len': minlength})
            raise ExtractionError(message)
    return val


@managedprops('ascii')
def ascii_extractor(widget, data):
    """Validate if a string is ASCII encoding.

    Only perform if ``ascii`` property evaludates to True.

    Properties:

    ``ascii``
        Flag  ascii check should perform.
    """
    val = data.extracted
    if val is UNSET:
        return val
    if not attr_value('ascii', widget, data, False):
        return val
    try:
        val.encode('ascii')
    except UnicodeEncodeError:
        message = _(u'ascii_extractor_error',
                    default=u'Input contains illegal characters.')
        raise ExtractionError(message)
    return val


LOWER_CASE_RE = r'(?=.*[a-z])'
UPPER_CASE_RE = r'(?=.*[A-Z])'
DIGIT_RE = r'(?=.*[\d])'
SPECIAL_CHAR_RE = r'(?=.*[\W])'
RE_PASSWORD_ALL = [
    LOWER_CASE_RE,
    UPPER_CASE_RE,
    DIGIT_RE,
    SPECIAL_CHAR_RE]
PASSWORD_NOCHANGE_VALUE = '_NOCHANGE_'


@managedprops('strength', 'weak_password_message')
def password_extractor(widget, data):
    """Extract and validate password input.

    If extracted password is unchanged, return ``UNSET``. Consider this when
    reading from password widgets!

    This extractor provides a strength check. It only performs if ``strenght``
    property is set. Strength check is done by four rules:
        - input contains lowercase character
        - input contains uppercase character
        - input contains digit
        - input contains special character.

    Properties:

    ``strength``
        Integer value <= 4. Define how many rules must apply to consider a
        password valid.
    """
    val = data.extracted
    if val == PASSWORD_NOCHANGE_VALUE:
        return UNSET
    if val is UNSET:
        return val
    required_strength = attr_value('strength', widget, data, 0)
    if required_strength <= 0:
        return val
    if required_strength > len(RE_PASSWORD_ALL):
        required_strength = len(RE_PASSWORD_ALL)
    strength = 0
    for reg_exp in RE_PASSWORD_ALL:
        if re.match(reg_exp, val):
            strength += 1
    if strength < required_strength:
        error = attr_value('weak_password_message', widget, data)
        raise ExtractionError(error)
    return val


def _pwd_value(widget, data):
    if data.extracted is not UNSET:
        return data.extracted
    if data.value is not UNSET \
       and data.value is not None:
        return PASSWORD_NOCHANGE_VALUE
    return attr_value('default', widget, data)


@managedprops('data', 'title', 'size', 'disabled', 'autofocus',
              'placeholder', 'autocomplete', *css_managed_props)
def password_edit_renderer(widget, data):
    """Render password widget.
    """
    tag = data.tag
    input_attrs = input_attributes_common(widget, data)
    input_attrs['type'] = 'password'
    input_attrs['value'] = _pwd_value(widget, data)
    return tag('input', **input_attrs)


@managedprops('displayplaceholder')
def password_display_renderer(widget, data):
    value = _pwd_value(widget, data)
    if value == PASSWORD_NOCHANGE_VALUE:
        return attr_value('displayplaceholder', widget, data)
    return u''


factory.register(
    'password',
    extractors=[
        generic_extractor,
        generic_required_extractor,
        generic_emptyvalue_extractor,
        minlength_extractor,
        ascii_extractor,
        password_extractor
    ],
    edit_renderers=[password_edit_renderer],
    display_renderers=[password_display_renderer])

factory.doc['blueprint']['password'] = """\
Password blueprint.

The password is never rendered to markup, instead
``yafowil.common.PASSWORD_NOCHANGE_VALUE`` is set as ``value`` property on
dom element. See ``yafowil.common.password_extractor`` for details on
password extraction.
"""

factory.defaults['password.required_class'] = 'required'

factory.defaults['password.default'] = ''

factory.defaults['password.class'] = 'password'

factory.defaults['password.minlength'] = -1
factory.doc['props']['password.size'] = """\
Maximum length of password.
"""

factory.doc['props']['password.minlength'] = """\
Minimum length of password.
"""

factory.defaults['password.ascii'] = False
factory.doc['props']['password.ascii'] = """\
Flag ascii check should performed.
"""

factory.defaults['password.strength'] = -1
factory.doc['props']['password.strength'] = """\
Integer value <= 4. Define how many rules must apply to consider a password
valid.
"""

factory.defaults['weak_password_message'] = _('weak_password_message',
                                              default=u'Password too weak')
factory.doc['props']['password.weak_password_message'] = """\
Message shown if password is not strong enough.
"""

factory.defaults['password.displayplaceholder'] = u'*' * 8
factory.doc['props']['password.displayplaceholder'] = """\
Placeholder shown in display mode if password was set.
"""

factory.defaults['password.persist'] = True


###############################################################################
# checkbox
###############################################################################

@managedprops('format')
def checkbox_extractor(widget, data):
    """Extracts data from a single input with type checkbox.
    """
    if '{0}-exists'.format(widget.dottedpath) not in data.request:
        return UNSET
    fmt = attr_value('format', widget, data)
    if fmt == 'bool':
        return widget.dottedpath in data.request
    elif fmt == 'string':
        return data.request.get(widget.dottedpath, '')
    raise ValueError(
        "Checkbox widget has invalid format '{0}' set".format(fmt)
    )


@managedprops('data', 'title', 'size', 'disabled', 'autofocus',
              'format', 'disabled', 'checked', 'with_label', *css_managed_props)
def checkbox_edit_renderer(widget, data):
    tag = data.tag
    input_attrs = input_attributes_common(widget, data)
    input_attrs['type'] = 'checkbox'
    checked = attr_value('checked', widget, data)
    if checked is not None:
        if checked:
            input_attrs['checked'] = 'checked'
    else:
        input_attrs['checked'] = input_attrs['value'] and 'checked' or None
    if attr_value('format', widget, data) == 'bool':
        input_attrs['value'] = ''
    checkbox = tag('input', **input_attrs)
    if attr_value('with_label', widget, data):
        checkbox += tag(
            'label',
            '&nbsp;',
            for_=cssid(widget, 'input'),
            class_='checkbox_label'
        )
    input_attrs = {
        'type': 'hidden',
        'value': 'checkboxexists',
        'name_': "{0}-exists".format(widget.dottedpath),
        'id': cssid(widget, 'checkboxexists'),
    }
    exists_marker = tag('input', **input_attrs)
    return checkbox + exists_marker


@managedprops('class', 'format', 'vocabulary', 'display_proxy')
def checkbox_display_renderer(widget, data):
    """Generic display renderer to render a value.
    """
    value = fetch_value(widget, data)
    fmt = attr_value('format', widget, data)
    if fmt == 'string' and bool(value):
        content = value
    else:
        vocab = dict(vocabulary(attr_value('vocabulary', widget, data, [])))
        # XXX: value might be 'True' Looks odd.
        content = vocab[bool(value)]
        if data.tag.translate:
            content = data.tag.translate(content)
    attrs = {
        'id': cssid(widget, 'display'),
        'class_': 'display-{0}'.format(attr_value('class', widget, data))
    }
    if attr_value('display_proxy', widget, data):
        widget.attrs['type'] = 'hidden'
        if fmt == 'string':
            input_attrs = input_attributes_common(widget, data, value=value)
            content += data.tag('input', **input_attrs)
        elif bool(value):
            input_attrs = input_attributes_common(widget, data, value='')
            content += data.tag('input', **input_attrs)
        del widget.attrs['type']
        input_attrs = {
            'type': 'hidden',
            'value': 'checkboxexists',
            'name_': "{0}-exists".format(widget.dottedpath),
            'id': cssid(widget, 'checkboxexists'),
        }
        content += data.tag('input', **input_attrs)
    return data.tag('div', content, **attrs)


factory.register(
    'checkbox',
    extractors=[
        checkbox_extractor,
        generic_required_extractor
    ],
    edit_renderers=[checkbox_edit_renderer],
    display_renderers=[checkbox_display_renderer])

factory.doc['blueprint']['checkbox'] = """\
Single checkbox blueprint.
"""

factory.defaults['checkbox.default'] = False

factory.defaults['checkbox.format'] = 'bool'
factory.doc['props']['checkbox.format'] = """\
Data-type of the extracted value. One out of ``bool`` or ``string``.
"""

factory.defaults['checkbox.class'] = 'checkbox'

factory.defaults['checkbox.disabled'] = False
factory.doc['props']['checkbox.disabled'] = """\
Flag whether checkbox is disabled.
"""

factory.defaults['checkbox.checked'] = None
factory.doc['props']['checkbox.checked'] = """\
Set 'checked' attribute explicit. If not given, compute by value.
"""

factory.defaults['checkbox.vocabulary'] = {
    True: _('yes', default=u'Yes'),
    False: _('no', default=u'No'),
    UNSET: _('unset', default=u'Unset'),  # XXX: never used right now?
}

factory.doc['props']['checkbox.vocabulary'] = """\
In display mode and if ```bool``` is set to ```True``` this mapping will be
used for display of the value. Expected keys are ```True```, ```False``` and
```UNSET```.
"""

factory.defaults['checkbox.with_label'] = False
factory.doc['props']['checkbox.with_label'] = """\
Render empty label tag after visible checkbox in order to make checkbox UI
customizable via CSS like so::

    input.large_checkbox {
        display: none;
    }
    input.large_checkbox + label {
        width: 59px;
        height: 60px;
        background: url('/checkbox_large.png');
    }
    input.large_checkbox:checked + label {
        background: url('/checkbox_large_selected.png');
    }
"""

factory.defaults['checkbox.required_class'] = 'required'

factory.defaults['checkbox.persist'] = True


###############################################################################
# select
###############################################################################

@managedprops('multivalued', 'disabled')
def select_extractor(widget, data):
    extracted = generic_extractor(widget, data)
    multivalued = attr_value('multivalued', widget, data)
    exists_marker = '{0}-exists'.format(widget.dottedpath)
    if extracted is UNSET and exists_marker in data.request:
        if multivalued:
            extracted = []
        else:
            extracted = ''
    if extracted is UNSET:
        return extracted
    if multivalued and isinstance(extracted, STR_TYPE):
        extracted = [extracted]
    disabled = widget.attrs.get('disabled', False)
    if not disabled:
        return extracted
    if not multivalued:
        return data.value
    disabled_items = disabled is True and data.value or disabled
    if isinstance(disabled_items, STR_TYPE):
        disabled_items = [disabled_items]
    for item in disabled_items:
        if item in extracted and item not in data.value:
            extracted.remove(item)
        elif item not in extracted and item in data.value:
            extracted.append(item)
    return extracted


def select_exists_marker(widget, data):
    tag = data.tag
    attrs = {
        'type': 'hidden',
        'value': 'exists',
        'name_': "{0}-exists".format(widget.dottedpath),
        'id': cssid(widget, 'exists'),
    }
    return tag('input', **attrs)


def select_edit_renderer_props(widget, data):
    value = fetch_value(widget, data)
    multivalued = attr_value('multivalued', widget, data)
    if isinstance(value, STR_TYPE) or not hasattr(value, '__iter__'):
        value = [value]
    datatype = widget.attrs.get('datatype', None)
    if datatype:
        value = convert_values_to_datatype(value, datatype)
    emptyvalue = attr_value('emptyvalue', widget, data, EMPTY_VALUE)
    if not multivalued and len(value) > 1:
        raise ValueError(u'Multiple values for single selection.')
    disabled = attr_value('disabled', widget, data, False)
    return value, multivalued, datatype, emptyvalue, disabled


def select_block_edit_renderer(widget, data, custom_attrs={}):
    value, multivalued, datatype, emptyvalue, disabled = \
        select_edit_renderer_props(widget, data)
    optiontags = []
    vocab = attr_value('vocabulary', widget, data, [])
    for key, term in vocabulary(vocab):
        vval = key
        if datatype:
            vval = convert_value_to_datatype(
                key,
                datatype,
                empty_value=emptyvalue
            )
        key = '' if key in [None, UNSET] else key
        attrs = {
            'selected': 'selected' if vval in value else None,
            'value': key,
            'id': cssid(widget, 'input', key),
        }
        if disabled and disabled is not True and vval in disabled:
            attrs['disabled'] = 'disabled'
        optiontags.append(data.tag('option', term, **attrs))
    autofocus = \
        attr_value('autofocus', widget, data) and 'autofocus' or None
    required = attr_value('required', widget, data) and 'required' or None
    block_class = attr_value('block_class', widget, data)
    select_attrs = {
        'name_': widget.dottedpath,
        'id': cssid(widget, 'input'),
        'title': attr_value('title', widget, data) or None,
        'class_': cssclasses(widget, data, additional=[block_class]),
        'multiple': multivalued and 'multiple' or None,
        'size': attr_value('size', widget, data) or None,
        'placeholder': attr_value('placeholder', widget, data) or None,
        'autofocus': autofocus,
        'required': required,
    }
    select_attrs.update(as_data_attrs(attr_value('data', widget, data)))
    select_attrs.update(custom_attrs)
    if disabled is True:
        select_attrs['disabled'] = 'disabled'
    if not optiontags:
        optiontags = [' ']
    rendered = data.tag('select', *optiontags, **select_attrs)
    if multivalued:
        attrs = {
            'type': 'hidden',
            'value': 'exists',
            'name_': '{0}-exists'.format(widget.dottedpath),
            'id': cssid(widget, 'exists'),
        }
        rendered = select_exists_marker(widget, data) + rendered
    return rendered


def select_cb_edit_renderer(widget, data, custom_attrs={}):
    value, multivalued, datatype, emptyvalue, disabled = \
        select_edit_renderer_props(widget, data)
    tags = []
    label_pos = attr_value('listing_label_position', widget, data)
    if label_pos == 'inner':
        # deprecated, use explicit inner-after or inner-before
        label_pos = 'inner-after'
    listing_tag = attr_value('listing_tag', widget, data)
    item_tag = listing_tag == 'div' and 'div' or 'li'
    if multivalued:
        tagtype = 'checkbox'
        wrapper_class = attr_value('checkbox_wrapper_class', widget, data)
        label_class = attr_value('checkbox_label_class', widget, data)
        input_class_additional = attr_value('checkbox_input_class', widget, data)
        # B/C deprecated as of yafowil 2.2
        if not label_class:
            label_class = attr_value('label_checkbox_class', widget, data)
    else:
        tagtype = 'radio'
        wrapper_class = attr_value('radio_wrapper_class', widget, data)
        label_class = attr_value('radio_label_class', widget, data)
        input_class_additional = attr_value('radio_input_class', widget, data)
        # B/C deprecated as of yafowil 2.2
        if not label_class:
            label_class = attr_value('label_radio_class', widget, data)
    vocab = attr_value('vocabulary', widget, data, [])
    for key, term in vocabulary(vocab):
        vval = key
        if datatype:
            vval = convert_value_to_datatype(
                key,
                datatype,
                empty_value=emptyvalue
            )
        key = '' if key in [None, UNSET] else key
        input_attrs = {
            'type': tagtype,
            'value': key,
            'checked': 'checked' if vval in value else None,
            'name_': widget.dottedpath,
            'id': cssid(widget, 'input', key),
            'class_': cssclasses(
                widget,
                data,
                additional=[input_class_additional]
            ),
        }
        if (disabled and disabled is not True and vval in disabled) \
           or disabled is True:
            input_attrs['disabled'] = 'disabled'
        inputtag = data.tag('input', **input_attrs)
        label_attrs = dict(for_=input_attrs['id'], _class=label_class)
        item = generic_positional_rendering_helper(
            'label', term, label_attrs, inputtag, label_pos, data.tag)
        item_wrapper = data.tag(item_tag, item, **{
            'id': cssid(widget, tagtype, key),
            'class': wrapper_class,
        })
        tags.append(item_wrapper)
    wrapper_attrs = {'id': cssid(widget, tagtype, 'wrapper')}
    wrapper_attrs.update(as_data_attrs(attr_value('data', widget, data)))
    wrapper_attrs.update(custom_attrs)
    taglisting = data.tag(listing_tag, *tags, **wrapper_attrs)
    return select_exists_marker(widget, data) + taglisting


@managedprops('data', 'title', 'format', 'vocabulary', 'multivalued',
              'disabled', 'listing_label_position', 'listing_tag', 'size',
              'block_class', 'autofocus', 'placeholder', 'datatype',
              'emptyvalue', 'checkbox_wrapper_class', 'checkbox_label_class',
              'checkbox_input_class', 'radio_wrapper_class',
              'radio_label_class', 'radio_input_class', *css_managed_props)
def select_edit_renderer(widget, data, custom_attrs={}):
    if attr_value('format', widget, data) == 'block':
        return select_block_edit_renderer(
            widget,
            data,
            custom_attrs=custom_attrs
        )
    return select_cb_edit_renderer(widget, data, custom_attrs=custom_attrs)


@managedprops('data', 'template', 'class', 'multivalued')
def select_display_renderer(widget, data):
    value = fetch_value(widget, data)
    if type(value) in ITER_TYPES and not value:
        value = u''
    multivalued = attr_value('multivalued', widget, data)
    vocab = dict(attr_value('vocabulary', widget, data, []))
    if not multivalued or not value:
        value = vocab.get(value, value)
        if data.tag.translate:
            value = data.tag.translate(value)
        return generic_display_renderer(widget, data, value=value)
    attrs = {
        'id': cssid(widget, 'display'),
        'class_': 'display-{0}'.format(attr_value('class', widget, data))
    }
    attrs.update(as_data_attrs(attr_value('data', widget, data)))
    content = u''
    if multivalued and isinstance(value, STR_TYPE):
        value = [value]
    for key in value:
        content += data.tag('li', vocab.get(key, key))
    return data.tag('ul', content, **attrs)


factory.register(
    'select',
    extractors=[
        select_extractor,
        generic_required_extractor,
        generic_emptyvalue_extractor,
        generic_datatype_extractor,
    ],
    edit_renderers=[select_edit_renderer],
    display_renderers=[
        select_display_renderer,
        display_proxy_renderer
    ])

factory.doc['blueprint']['select'] = """\
Selection Blueprint. Single selection as dropdown or radio-buttons. Multiple
selection as selection-list or as checkboxes.
"""

factory.defaults['select.multivalued'] = None
factory.doc['props']['select.multivalued'] = """\
Flag whether multiple items can be selected.
"""

factory.defaults['select.size'] = None
factory.doc['props']['select.size'] = """\
Size of input if multivalued and format 'block'.
"""

# maybe callable returning '' for single select and [] for multi select
factory.defaults['select.default'] = UNSET

factory.defaults['select.format'] = 'block'
factory.doc['props']['select.format'] = """\
Every value except 'block' results in either a list of radio buttons or
checkboxes depending on the 'multivalued' property.
"""

factory.defaults['select.class'] = 'select'

factory.defaults['select.block_class'] = None
factory.doc['props']['select.block_class'] = """\
CSS class to render on selection if block format.
"""

factory.defaults['select.checkbox_wrapper_class'] = None
factory.doc['props']['select.checkbox_wrapper_class'] = """\
CSS class to render on checkbox wrapper.
"""

factory.defaults['select.checkbox_label_class'] = None
factory.doc['props']['select.checkbox_label_class'] = """\
CSS class to render on checkbox labels.
"""

factory.defaults['select.label_checkbox_class'] = None
factory.doc['props']['select.label_checkbox_class'] = """\
CSS class to render on checkbox labels.

This property is deprecated and will be remove as of yafowil 2.2. Use
``checkbox_label_class`` instead.
"""

factory.defaults['select.checkbox_input_class'] = None
factory.doc['props']['select.checkbox_input_class'] = """\
CSS class to render on checkbox input tag.
"""

factory.defaults['select.radio_wrapper_class'] = None
factory.doc['props']['select.radio_wrapper_class'] = """\
CSS class to render on radio button wrapper.
"""

factory.defaults['select.radio_label_class'] = None
factory.doc['props']['select.radio_label_class'] = """\
CSS class to render on radio button labels.
"""

factory.defaults['select.label_radio_class'] = None
factory.doc['props']['select.label_radio_class'] = """\
CSS class to render on radio button labels.

This property is deprecated and will be remove as of yafowil 2.2. Use
``radio_label_class`` instead.
"""

factory.defaults['select.radio_input_class'] = None
factory.doc['props']['select.radio_input_class'] = """\
CSS class to render on radio button input tag.
"""

factory.defaults['select.listing_tag'] = 'div'
factory.doc['props']['select.listing_tag'] = """\
Desired rendering tag for selection if selection format is 'single'. Valid
values are 'div' and 'ul'.
"""

factory.defaults['select.listing_label_position'] = 'inner-after'
factory.doc['props']['select.listing_label_position'] = """\
Label position if format is 'single'. Behaves the same way as label widget
position property.
"""

factory.doc['props']['select.vocabulary'] = """\
Vocabulary to be used for the selection list. Expects a dict-like or an
iterable or a callable which returns one of both first. An iterable can consist
out of strings or out of tuples with ``(key, value)``. The items in the result
list are in the same order like the vocabulary.
"""

factory.doc['props']['select.disabled'] = """\
Disables the whole widget or single selections. To disable the whole widget
set the value to 'True'. To disable single selection pass a iterable of keys to
disable, i.e. ``['foo', 'baz']``. Defaults to False.
"""

factory.defaults['select.persist'] = True


###############################################################################
# file
###############################################################################

def file_extractor(widget, data):
    """Return a dict with following keys:

    mimetype
        Mimetype of file.
    headers
        rfc822.Message instance.
    original
        Original file handle from underlying framework.
    file
        File descriptor containing the data.
    filename
        File name.
    action
        widget flags 'new', 'keep', 'replace', 'delete'
    """
    name = widget.dottedpath
    if name not in data.request:
        return UNSET
    if not '{0}-action'.format(name) in data.request:
        value = data.request[name]
        if value:
            value['action'] = 'new'
        return value
    value = data.value
    action = value['action'] = data.request.get(
        '{0}-action'.format(name),
        'keep'
    )
    if action == 'delete':
        value['file'] = UNSET
    elif action == 'replace':
        new_val = data.request[name]
        if new_val:
            value = new_val
            value['action'] = 'replace'
        else:
            value['action'] = 'keep'
    return value


@managedprops('accept', 'placeholder', 'autofocus',
              'required', *css_managed_props)
def input_file_edit_renderer(widget, data):
    tag = data.tag
    input_attrs = input_attributes_common(widget, data, excludes=['value'])
    input_attrs['type'] = 'file'
    if attr_value('accept', widget, data):
        input_attrs['accept'] = attr_value('accept', widget, data)
    return tag('input', **input_attrs)


def convert_bytes(value):
    value = float(value)
    if value >= 1099511627776:
        terabytes = value / 1099511627776
        size = '{0:.2f}T'.format(terabytes)
    elif value >= 1073741824:
        gigabytes = value / 1073741824
        size = '{0:.2f}G'.format(gigabytes)
    elif value >= 1048576:
        megabytes = value / 1048576
        size = '{0:.2f}M'.format(megabytes)
    elif value >= 1024:
        kilobytes = value / 1024
        size = '{0:.2f}K'.format(kilobytes)
    else:
        size = '{0:.2f}b'.format(value)
    return size


def input_file_display_renderer(widget, data):
    tag = data.tag
    value = data.value
    attrs = {
        'class': cssclasses(widget, data),
    }
    attrs.update(as_data_attrs(attr_value('data', widget, data)))
    if not value:
        no_file_message = _('no_file', default=u'No file')
        return tag('div', no_file_message, **attrs)
    file_val = value['file']
    size = convert_bytes(len(file_val.read()))
    file_val.seek(0)
    unknown_message = _('unknown', default=u'Unknown')
    filename_message = _('filename', default=u'Filename: ')
    mimetype_message = _('mimetype', default=u'Mimetype: ')
    size_message = _('size', default=u'Size: ')
    filename = value.get('filename', unknown_message)
    mimetype = value.get('mimetype', unknown_message)
    return tag('div',
               tag('ul',
                   tag('li', tag('strong', filename_message), filename),
                   tag('li', tag('strong', mimetype_message), mimetype),
                   tag('li', tag('strong', size_message), size)),
               **attrs)


@managedprops(*css_managed_props)
def file_options_renderer(widget, data):
    if data.value in [None, UNSET, '']:
        return data.rendered
    tag = data.tag
    if data.request:
        value = [
            data.request.get('{0}-action'.format(widget.dottedpath), 'keep')
        ]
    else:
        value = ['keep']
    tags = []
    vocab = attr_value('vocabulary', widget, data, [])
    for key, term in vocabulary(vocab):
        attrs = {
            'type': 'radio',
            'value': key,
            'checked': (key in value) and 'checked' or None,
            'name_': '{0}-action'.format(widget.dottedpath),
            'id': cssid(widget, 'input', key),
            'class_': cssclasses(widget, data),
        }
        taginput = tag('input', **attrs)
        text = tag('span', term)
        tags.append(tag('div', taginput, text,
                        **{'id': cssid(widget, 'radio', key)}))
    return data.rendered + u''.join(tags)


factory.register(
    'file',
    extractors=[
        file_extractor,
        generic_required_extractor
    ],
    edit_renderers=[
        input_file_edit_renderer,
        file_options_renderer
    ],
    display_renderers=[input_file_display_renderer])

factory.doc['blueprint']['file'] = """\
A basic file upload blueprint.
"""

factory.defaults['file.accept'] = None
factory.doc['props']['file.accept'] = """\
Accepted mimetype.
"""

factory.defaults['file.vocabulary'] = [
    ('keep', _('file_keep', default=u'Keep Existing file')),
    ('replace', _('file_replace', default=u'Replace existing file')),
    ('delete', _('file_delete', default=u'Delete existing file')),
]


###############################################################################
# submit
###############################################################################

@managedprops('tag_type', 'label', 'class', 'action', 'handler',
              'next', 'skip', 'expression')
def submit_renderer(widget, data):
    expression = attr_value('expression', widget, data)
    if not expression:
        return u''
    tag = data.tag
    input_attrs = input_attributes_common(widget, data)
    input_attrs['type'] = 'submit'
    input_attrs['name_'] = attr_value('action', widget, data) \
        and 'action.{0}'.format(widget.dottedpath)
    input_attrs['value'] = attr_value('label', widget, data, widget.name)
    return tag(attr_value('tag_type', widget, data), **input_attrs)


factory.register(
    'submit',
    edit_renderers=[submit_renderer],
    display_renderers=[empty_display_renderer])

factory.doc['blueprint']['submit'] = """\
Submit action inside the form
"""

factory.doc['props']['submit.label'] = """\
Label of the submit.
"""

factory.defaults['submit.tag_type'] = 'input'
factory.doc['props']['submit.tag_type'] = """\
Define the type of tag that will be rendered.
"""

factory.defaults['submit.expression'] = True
factory.doc['props']['submit.expression'] = """\
Flag or expression callable whether this action is available to the user
or not.
"""

factory.defaults['submit.action'] = True
factory.doc['props']['submit.action'] = """\
Marks this widget as an action. One out of ``True`` or ``False``.
"""

factory.defaults['submit.skip'] = False
factory.doc['props']['submit.skip'] = """\
Skips action and only perform next. One out of ``True`` or ``False``.
"""

factory.doc['props']['submit.handler'] = """\
Handler is a callable which get called if this action performs. It expects two
parameters: ``widget``, ``data``.
"""

factory.doc['props']['submit.next'] = """\
Next is a callable expected to return the web address. It expects a request as
the only parameter.
"""

factory.defaults['submit.disabled'] = False
factory.doc['props']['submit.disabled'] = """\
Flag the submit field as disabled.
"""


###############################################################################
# button
###############################################################################

@managedprops(
    'text', 'action', 'handler', 'next', 'skip', 'type',
    'expression', 'form', 'formaction', 'formenctype', 'formmethod',
    'formnovalidate', 'formtarget', 'autofocus', 'disabled', 'class',
    'class_add', 'accesskey')
def button_renderer(widget, data):
    expression = attr_value('expression', widget, data)
    if not expression:
        return u''
    tag = data.tag
    input_attrs = input_attributes_common(widget, data)
    input_attrs['type'] = attr_value('type', widget, data)
    input_attrs['form'] = attr_value('form', widget, data)
    input_attrs['formaction'] = attr_value('formaction', widget, data)
    input_attrs['formenctype'] = attr_value('formenctype', widget, data)
    input_attrs['formmethod'] = attr_value('formmethod', widget, data)
    input_attrs['formnovalidate'] = attr_value('formnovalidate', widget, data)
    input_attrs['formtarget'] = attr_value('formtarget', widget, data)
    input_attrs['accesskey'] = attr_value('accesskey', widget, data)
    if input_attrs['type'] == 'submit':
        # only 'submit' type sends data to the server
        if attr_value('action', widget, data):
            # only if an server side action is set this makes sense:
            input_attrs['name_'] = 'action.{0}'.format(widget.dottedpath)
        else:
            # otherwise check for a custom name
            input_attrs['name_'] = attr_value('name', widget, data)
    text = attr_value('text', widget, data)
    return tag("button", text, **input_attrs)


factory.register(
    'button',
    edit_renderers=[button_renderer],
    display_renderers=[empty_display_renderer])

factory.defaults['button.type'] = 'submit'

factory.doc['blueprint']['button'] = """\
Represents a clickable button, used to submit forms or anywhere in a document for
accessible, standard button functionality.
"""

factory.doc['props']['button.text'] = """\
The content of the button element.
"""

# parts of the docs are copied over from
# https://developer.mozilla.org/de/docs/Web/HTML/Element/button

factory.doc['props']['button.form'] = """\
The <form> element to associate the button with (its form owner).
The value of this attribute must be the id of a <form> in the same document.
(If this attribute is not set, the <button> is associated with its ancestor
<form> element, if any.)
"""

factory.doc['props']['button.formaction'] = """\
The URL that processes the information submitted by the button.
Overrides the action attribute of the button's form owner.
Does nothing if there is no form owner.
"""

factory.doc['props']['button.formenctype'] = """\
If the button is a submit button (it's inside/associated with a <form> and
doesn't have type="button"), specifies how to encode the form data that is
submitted. Possible values:

application/x-www-form-urlencoded: The default if the attribute is not used.

multipart/form-data: Use to submit <input> elements with their type attributes
set to file.

text/plain: Specified as a debugging aid; shouldn’t be used for real form
submission.

If this attribute is specified, it overrides the enctype attribute of the
button's form owner.
"""

factory.doc['props']['button.formmethod'] = """\
If the button is a submit button (it's inside/associated with a <form> and
doesn't have type="button"), this attribute specifies the HTTP method used
to submit the form. Possible values: POST or GET. If specified, this
attribute overrides the method attribute of the button's form owner.
"""

factory.doc['props']['button.formnovalidate'] = """\
If the button is a submit button, this Boolean attribute specifies that the
form is not to be validated when it is submitted. If this attribute is
specified, it overrides the novalidate attribute of the button's form owner.
"""

factory.doc['props']['button.formtarget'] = """\
If the button is a submit button, this attribute is a author-defined name or
standardized, underscore-prefixed keyword indicating where to display the
response from submitting the form. This is the name of, or keyword for,
a browsing context (a tab, window, or <iframe>). If this attribute is
specified, it overrides the target attribute of the button's form owner.
The following keywords have special meanings:

_self: Load the response into the same browsing context as the current one.
This is the default if the attribute is not specified.

_blank: Load the response into a new unnamed browsing context — usually a new
tab or window, depending on the user’s browser settings.

_parent: Load the response into the parent browsing context of the current one.
If there is no parent, this option behaves the same way as _self.

_top: Load the response into the top-level browsing context (that is, the
browsing context that is an ancestor of the current one, and has no parent).
If there is no parent, this option behaves the same way as _self.
"""

factory.defaults['button.expression'] = True
factory.doc['props']['button.expression'] = """\
Flag or expression callable whether this action is available to the user
or not.
"""

factory.defaults['button.action'] = True
factory.doc['props']['button.action'] = """\
Marks this widget as an action. One out of ``True`` or ``False``.
"""

factory.defaults['button.skip'] = False
factory.doc['props']['button.skip'] = """\
Skips action and only perform next. One out of ``True`` or ``False``.
"""

factory.doc['props']['button.handler'] = """\
Handler is a callable which get called if this action performs. It expects two
parameters: ``widget``, ``data``.
"""

factory.doc['props']['button.next'] = """\
Next is a callable expected to return the web address. It expects a request as
the only parameter.
"""

factory.defaults['button.disabled'] = False
factory.doc['props']['button.disabled'] = """\
Flag the button field as disabled.
"""

factory.defaults['button.autofocus'] = False
factory.doc['props']['button.autofocus'] = """\
Boolean attribute specifies that the button should have input focus when the
page loads. Only one element in a document can have this attribute.
"""

factory.defaults['button.autocomplete'] = False
factory.doc['props']['button.autocomplete'] = """\
This attribute on a <button> is nonstandard and Firefox-specific.
Unlike other browsers, Firefox persists the dynamic disabled state of a
<button> across page loads. Setting autocomplete="off" on the button
disables this feature
"""


###############################################################################
# email
###############################################################################

EMAIL_RE = r'^[a-zA-Z0-9\._\-]+@[a-zA-Z0-9\._\-]+.[a-zA-Z0-9]{2,6}$'


def email_extractor(widget, data):
    val = data.extracted
    if not val:
        return val
    email_re = EMAIL_RE \
        if isinstance(val, UNICODE_TYPE) \
        else EMAIL_RE.encode()
    if not re.match(email_re, val):
        message = _('email_address_not_valid',
                    default=u'Input not a valid email address.')
        raise ExtractionError(message)
    return val


factory.register(
    'email',
    extractors=[
        generic_extractor,
        generic_required_extractor,
        generic_emptyvalue_extractor,
        generic_datatype_extractor,
        email_extractor
    ],
    edit_renderers=[input_generic_renderer],
    display_renderers=[
        generic_display_renderer,
        display_proxy_renderer
    ])

factory.doc['blueprint']['email'] = """\
Email (HTML5) input blueprint.
"""

factory.defaults['email.type'] = 'email'

factory.defaults['email.default'] = ''

factory.defaults['email.required_class'] = 'required'

factory.defaults['email.class'] = 'email'

factory.defaults['email.persist'] = True

factory.defaults['email.allowed_datatypes'] = [
    BYTES_TYPE, UNICODE_TYPE,
]

###############################################################################
# url
###############################################################################

URL_RE = r'^(ftp|http|https):\/\/(\w+:{0,1}\w*@)?(\S+)(:[0-9]+)?(\/|\/([\w#!:'
URL_RE += r'.?+=&%@!\-\/]))?$'


def url_extractor(widget, data):
    val = data.extracted
    if not val:
        return val
    if not re.match(URL_RE, val is not UNSET and val or ''):
        message = _('web_address_not_valid',
                    default=u'Input not a valid web address.')
        raise ExtractionError(message)
    return val


factory.register(
    'url',
    extractors=[
        generic_extractor,
        generic_required_extractor,
        generic_emptyvalue_extractor,
        url_extractor
    ],
    edit_renderers=[input_generic_renderer],
    display_renderers=[
        generic_display_renderer,
        display_proxy_renderer
    ])

factory.doc['blueprint']['url'] = """\
URL aka web address (HTML5) input blueprint.
"""

factory.defaults['url.type'] = 'url'

factory.defaults['url.default'] = ''

factory.defaults['url.required_class'] = 'required'

factory.defaults['url.class'] = 'url'

factory.defaults['url.persist'] = True


###############################################################################
# search
###############################################################################

factory.register(
    'search',
    extractors=[
        generic_extractor,
        generic_required_extractor,
        generic_emptyvalue_extractor,
    ],
    edit_renderers=[input_generic_renderer],
    display_renderers=[
        generic_display_renderer,
        display_proxy_renderer
    ])

factory.doc['blueprint']['search'] = """\
Search blueprint (HTML5).
"""

factory.defaults['search.type'] = 'search'

factory.defaults['search.default'] = ''

factory.defaults['search.required_class'] = 'required'

factory.defaults['search.class'] = 'search'


###############################################################################
# number
###############################################################################

@managedprops('min', 'max', 'step')
def number_extractor(widget, data):
    val = data.extracted
    if val is UNSET:
        return val
    min_val = attr_value('min', widget, data)
    if min_val is not None and val < min_val:
        message = _('input_number_minimum_value',
                    default=u'Value has to be at minimum ${min}.',
                    mapping={'min': min_val})
        raise ExtractionError(message)
    max_val = attr_value('max', widget, data)
    if max_val is not None and val > max_val:
        message = _('input_number_maximum_value',
                    default=u'Value has to be at maximum ${max}.',
                    mapping={'max': max_val})
        raise ExtractionError(message)
    step = attr_value('step', widget, data)
    if step:
        minimum = min_val or 0
        if (val - minimum) % step:
            if minimum:
                message = _(
                    'input_number_step_and_minimum_value',
                    default=u'Value ${val} has to be in stepping of ${step} '
                            u'based on a floor value of ${minimum}',
                    mapping={
                        'val': val,
                        'step': step,
                        'minimum': minimum,
                    })
            else:
                message = _(
                    'input_number_step_value',
                    default=u'Value ${val} has to be in stepping of ${step}',
                    mapping={
                        'val': val,
                        'step': step,
                    })
            raise ExtractionError(message)
    return val


factory.register(
    'number',
    extractors=[
        generic_extractor,
        generic_required_extractor,
        generic_emptyvalue_extractor,
        generic_datatype_extractor,
        number_extractor,
    ],
    edit_renderers=[input_generic_renderer],
    display_renderers=[
        generic_display_renderer,
        display_proxy_renderer
    ])

factory.doc['blueprint']['number'] = """\
Number blueprint (HTML5).
"""

factory.defaults['number.type'] = 'number'

factory.defaults['number.default'] = ''

factory.defaults['number.emptyvalue'] = UNSET

factory.defaults['number.datatype'] = float
factory.doc['props']['number.datatype'] = """\
Callable for converting extracted value to output datatype. Allowed datatypes
are ``int`` and ``float``

``datatype`` can also be defined as string with value out of `'int'``,
``'integer'`` or ``'float'``.
"""

factory.defaults['number.allowed_datatypes'] = [
    int, float,
    # B/C
    'int', 'integer', 'float'
]

factory.defaults['number.min'] = None
factory.doc['props']['number.min'] = """\
Minimum value.
"""

factory.defaults['number.max'] = None
factory.doc['props']['number.max'] = """\
Maximum value.
"""

factory.defaults['number.step'] = None
factory.doc['props']['number.step'] = """\
Stepping value must be in.
"""

factory.defaults['number.required_class'] = 'required'

factory.defaults['number.class'] = 'number'

factory.defaults['number.persist'] = True


###############################################################################
# label
###############################################################################

@managedprops('position', 'label', 'for', *css_managed_props)
def label_renderer(widget, data):
    tag = data.tag
    label_text = attr_value('label', widget, data, widget.name)
    label_attrs = {
        'class_': cssclasses(widget, data)
    }
    if data.mode == 'edit':
        for_path = attr_value('for', widget, data)
        if for_path:
            for_widget = widget.root
            for name in for_path.split('.'):
                for_widget = for_widget[name]
            label_attrs['for_'] = cssid(for_widget, 'input')
        else:
            label_attrs['for_'] = cssid(widget, 'input')
        title = attr_value('title', widget, data)
        if title:
            label_attrs['title'] = title
    pos = attr_value('position', widget, data)
    if pos == 'inner':
        # deprecated, use explicit inner-after or inner-before
        pos = 'inner-before'
    rendered = data.rendered is not UNSET and data.rendered or u''
    return generic_positional_rendering_helper(
        'label', label_text, label_attrs, rendered, pos, tag
    )


factory.register(
    'label',
    edit_renderers=[label_renderer],
    display_renderers=[label_renderer])

factory.doc['blueprint']['label'] = """\
Label blueprint.
"""

factory.defaults['label.position'] = 'before'
factory.doc['props']['label.position'] = """\
Label can be rendered at 3 different positions: ``before`` or ``after`` the
prior rendered output or with ``inner`` it puts the prior rendered output
inside the label tag.
"""

factory.doc['props']['label.label'] = """\
Text to be displayed as a label.
"""

factory.defaults['label.for'] = None
factory.doc['props']['label.for'] = """\
Optional dottedpath of widget to be labled
"""


###############################################################################
# field
###############################################################################

@managedprops('witherror', *css_managed_props)
def field_renderer(widget, data):
    tag = data.tag
    div_attrs = {
        'id': cssid(widget, 'field'),
        'class_': cssclasses(widget, data)
    }
    witherror = attr_value('witherror', widget, data)
    if witherror and data.errors:
        div_attrs['class_'] += u' {0}'.format(witherror)
    return tag('div', data.rendered, **div_attrs)


factory.register(
    'field',
    edit_renderers=[field_renderer],
    display_renderers=[field_renderer])

factory.doc['blueprint']['field'] = """\
Renders a div with an class field around the prior rendered output. This is
supposed to be used for styling and grouping purposes.
"""

factory.defaults['field.class'] = 'field'

factory.defaults['field.witherror'] = None
factory.doc['props']['field.witherror'] = """\
Put the class given with this property on the div if an error happened.
"""


###############################################################################
# error
###############################################################################

@managedprops('tag', 'message_tag', 'message_class', 'position',
              'render_empty', *css_managed_props)
def error_renderer(widget, data):
    if not data.errors and not attr_value('render_empty', widget, data):
        return data.rendered
    tag = data.tag
    msgs = u''
    for error in data.errors:
        message_tag = attr_value('message_tag', widget, data)
        if message_tag:
            msgs += tag(
                message_tag,
                error.msg,
                class_=attr_value('message_class', widget, data)
            )
        else:
            msgs += tag.translate(error.msg)
    attrs = dict(class_=cssclasses(widget, data))
    elem_tag = attr_value('tag', widget, data)
    position = attr_value('position', widget, data)
    return generic_positional_rendering_helper(elem_tag, msgs,
                                               attrs, data.rendered,
                                               position, tag)


factory.register(
    'error',
    edit_renderers=[error_renderer],
    display_renderers=[empty_display_renderer])

factory.doc['blueprint']['error'] = """\
Renders a tag with an error-message and the prior rendered output.
"""

factory.defaults['error.class'] = 'error'

factory.defaults['error.tag'] = 'div'
factory.doc['props']['error.tag'] = """\
HTML tag to use to enclose all error messages.
"""

factory.defaults['error.render_empty'] = False
factory.doc['props']['error.render_empty'] = """\
Render tag even if there is no error message.
"""

factory.defaults['error.message_tag'] = 'div'
factory.doc['props']['error.message_tag'] = """\
HTML tag to use to enclose each error message.
"""

factory.defaults['error.message_class'] = 'errormessage'
factory.doc['props']['error.message_class'] = """\
CSS class to apply to inner message-tag.
"""

factory.defaults['error.position'] = 'inner-before'
factory.doc['props']['error.position'] = """\
Error can be rendered at 3 different positions: ``before``/ ``after`` the
prior rendered output or with ``inner-before``/ ``inner-after``  it puts the
prior rendered output inside the tag used for the error message (beofre or
after the message.
"""


###############################################################################
# help
###############################################################################

@managedprops('tag', 'help', 'position', 'render_empty', *css_managed_props)
def help_renderer(widget, data):
    render_empty = attr_value('render_empty', widget, data)
    help_val = attr_value('help', widget, data)
    if not render_empty and not help_val:
        return data.rendered
    tag = data.tag
    attrs = dict(class_=cssclasses(widget, data))
    elem_tag = attr_value('tag', widget, data)
    position = attr_value('position', widget, data)
    return generic_positional_rendering_helper(
        elem_tag, help_val, attrs, data.rendered, position, tag)


factory.register(
    'help',
    edit_renderers=[help_renderer],
    display_renderers=[empty_display_renderer])

factory.doc['blueprint']['help'] = """\
Renders a tag with an help-message and the prior rendered output.
"""

factory.defaults['help.class'] = 'help'

factory.defaults['help.tag'] = 'div'
factory.doc['props']['help.tag'] = """\
HTML tag to use to enclose all help messages.
"""

factory.defaults['help.help'] = ''
factory.doc['props']['help.help'] = """\
Help text.
"""

factory.defaults['help.render_empty'] = False
factory.doc['props']['help.render_empty'] = """\
Render tag even if there is no help message.
"""

factory.defaults['help.position'] = 'before'
factory.doc['props']['help.position'] = """\
Help can be rendered at 3 different positions: ``before``/ ``after`` the
prior rendered output or with ``inner-before``/ ``inner-after``  it puts the
prior rendered output inside the tag used for the help message (beofre or
after the message.
"""



from node.base import AttributedNode
from node.utils import UNSET
from odict import odict
from yafowil.base import factory
from yafowil.persistence import attribute_writer
from yafowil.persistence import node_attribute_writer
from yafowil.persistence import write_mapping_writer
from yafowil.tests import YafowilTestCase
import uuid
import yafowil.loader  # noqa


class TestPersistence(YafowilTestCase):

    def test_persistence(self):
        # Test ``attribute_writer``
        form = factory(
            'form',
            name='form',
            props={
                'persist_writer': attribute_writer
            })
        form['my_field'] = factory(
            'text',
            props={
                'persist': True
            })
        data = form.extract(request={
            'form.my_field': 'value'
        })
        self.assertEqual(data.persist_writer, attribute_writer)

        class AttributeModel(object):
            my_field = UNSET

        model = AttributeModel()
        data.write(model)
        self.assertEqual(model.my_field, 'value')

        # Test ``write_mapping_writer``
        form.attrs['persist_writer'] = write_mapping_writer
        data = form.extract(request={
            'form.my_field': 'value'
        })
        self.assertEqual(data.persist_writer, write_mapping_writer)

        model = dict()
        data.write(model)
        self.assertEqual(sorted(model.items()), [('my_field', 'value')])

        # Test ``node_attribute_writer``
        form.attrs['persist_writer'] = node_attribute_writer
        data = form.extract(request={
            'form.my_field': 'value'
        })
        self.assertEqual(data.persist_writer, node_attribute_writer)

        model = AttributedNode()
        data.write(model)
        self.assertEqual(sorted(model.attrs.items()), [('my_field', 'value')])

        # If ``recursiv`` flag is ``False``, don't dig into child data
        form.attrs['persist_writer'] = write_mapping_writer
        data = form.extract(request={
            'form.my_field': 'value'
        })
        self.checkOutput("""
        <RuntimeData form, value=<UNSET>, extracted=odict(...) at ...>
          <RuntimeData form.my_field, value=<UNSET>, extracted='value' at ...>
        """, data.treerepr())

        model = dict()
        data.write(model, recursiv=False)
        self.assertEqual(sorted(model.items()), [])

        # By default persitence ``target`` is taken from widget ``name``.
        # Widgets can persist to different ``target`` as well
        form['my_field'].attrs['persist_target'] = 'my_other_field'
        data = form.extract(request={
            'form.my_field': 'value'
        })
        model = dict()
        data.write(model)
        self.assertEqual(sorted(model.items()), [('my_other_field', 'value')])

        form['my_field'].attrs['persist_target'] = 'my_field'

        # We can override persist writers for specific widgets
        form['my_other_field'] = factory(
            'text',
            props={
                'persist': True,
                'persist_writer': attribute_writer
            })

        class MixedModel(dict):
            my_other_field = UNSET

        data = form.extract(request={
            'form.my_field': 'value',
            'form.my_other_field': 'other value'
        })
        model = MixedModel()
        data.write(model)

        self.assertEqual(model.my_other_field, 'other value')
        self.assertEqual(sorted(model.items()), [('my_field', 'value')])

        del form['my_other_field']

        # We can also persist extracted from compound
        form['compound'] = factory(
            'compound',
            props={
                'persist': True,
            })
        form['compound']['f1'] = factory(
            'text',
            props={
                'persist': False
            })
        form['compound']['f2'] = factory(
            'text',
            props={
                'persist': False
            })

        data = form.extract(request={
            'form.my_field': 'value',
            'form.compound.f1': 'f1',
            'form.compound.f2': 'f2'
        })
        model = dict()
        data.write(model)
        self.assertEqual(sorted(model.items()), [
            ('compound', odict([('f1', 'f1'), ('f2', 'f2')])),
            ('my_field', 'value')
        ])

        # We can call write on children of runtime data. ``writer`` must be
        # passed in if ``persist_writer`` is not provide via widget
        model = dict()
        with self.assertRaises(ValueError) as arc:
            data['my_field'].write(model)
        self.assertEqual(str(arc.exception), 'No persistence writer found')

        data['my_field'].write(model, writer=write_mapping_writer)
        self.assertEqual(sorted(model.items()), [('my_field', 'value')])

        # Structural children work as usual
        del form['compound'].attrs['persist']
        form['compound'].attrs['structural'] = True
        form['compound']['f1'].attrs['persist'] = True
        form['compound']['f2'].attrs['persist'] = True
        data = form.extract(request={
            'form.my_field': 'value',
            'form.compound.f1': 'f1',
            'form.compound.f2': 'f2'
        })
        model = dict()
        data.write(model)
        self.assertEqual(sorted(model.items()), [
            ('f1', 'f1'),
            ('f2', 'f2'),
            ('my_field', 'value')
        ])

        # Widgets with no persis flag set gets ignored
        form['compound']['f2'].attrs['persist'] = False
        data = form.extract(request={
            'form.my_field': 'value',
            'form.compound.f1': 'f1',
            'form.compound.f2': 'f2'
        })
        model = dict()
        data.write(model)
        self.assertEqual(
            sorted(model.items()),
            [('f1', 'f1'), ('my_field', 'value')]
        )

        form['compound']['f2'].attrs['persist'] = True

        # In conjunction with ``datatype`` and ``emptyvalue`` we have fancy
        # convenience for single model bound forms
        form['my_field'].attrs['datatype'] = int
        form['my_field'].attrs['emptyvalue'] = UNSET
        form['compound']['f1'].attrs['datatype'] = float
        form['compound']['f2'].attrs['datatype'] = uuid.UUID

        data = form.extract(request={
            'form.my_field': '',
            'form.compound.f1': '1.0',
            'form.compound.f2': 'c6b19794-dc9e-4a98-920d-182a9ec07b7a'
        })
        model = dict()
        data.write(model)
        self.assertEqual(sorted(model.items()), [
            ('f1', 1.0),
            ('f2', uuid.UUID('c6b19794-dc9e-4a98-920d-182a9ec07b7a')),
            ('my_field', UNSET)
        ])

        # When trying to persist data containing errors we get a runtime error
        data = form.extract(request={
            'form.my_field': '',
            'form.compound.f1': 'a',
            'form.compound.f2': 'c6b19794-dc9e-4a98-920d-182a9ec07b7a'
        })
        self.assertTrue(data.has_errors)

        with self.assertRaises(RuntimeError) as arc:
            data.write(dict())
        msg = 'Attempt to persist data which failed to extract'
        self.assertEqual(str(arc.exception), msg)



# -*- coding: utf-8 -*-
from node.utils import UNSET
from pkg_resources import iter_entry_points
from yafowil.compat import BYTES_TYPE
from yafowil.compat import IS_PY2
from yafowil.compat import LONG_TYPE
from yafowil.compat import STR_TYPE
from yafowil.compat import UNICODE_TYPE
import json
import logging
import re
import unicodedata
import uuid


class entry_point(object):
    """Decorator for yafowil entry points.
    """

    def __init__(self, order=0):
        self.order = order

    def __call__(self, ob):
        ob.order = self.order
        return ob


_yafowil_plugins = None


def get_plugins(ns=None):
    global _yafowil_plugins
    if _yafowil_plugins is None:
        _yafowil_plugins = list()
        for ep in iter_entry_points('yafowil.plugin'):
            cb = ep.load()
            _yafowil_plugins.append((ep, cb))
        _yafowil_plugins = sorted(
            _yafowil_plugins,
            key=lambda x: getattr(x[1], 'order', 0)
        )
    for ep, cb in _yafowil_plugins:
        if ns is not None and ep.name != ns:
            continue
        yield (ep, cb)


_plugin_names = dict()


def get_plugin_names(ns=None):
    if ns not in _plugin_names:
        _plugin_names[ns] = list(set(
            [ep.dist.project_name for ep, cb in get_plugins(ns=ns)]
        ))
    return _plugin_names[ns]


def get_example(example_name):
    for ep, cb in get_plugins(ns='example'):
        if ep.dist.project_name != example_name:
            continue
        info = cb()
        return info


def get_example_names():
    result = []
    for ep, cb in get_plugins(ns='example'):
        result.append(ep.dist.project_name)
    return result


def vocabulary(definition):
    """Convert different kinds of input into a list of bi-tuples, both strings.
    """
    if callable(definition):
        definition = definition()
    if isinstance(definition, STR_TYPE):
        return [(definition, definition), ]
    # dict-like
    if hasattr(definition, '__getitem__') and hasattr(definition, 'keys'):
        return [(_, definition[_]) for _ in definition.keys()]
    # iterable
    if hasattr(definition, '__iter__'):
        new_vocab = []
        for entry in definition:
            if isinstance(entry, STR_TYPE):
                # entry is a string
                new_vocab.append((entry, entry))
            elif hasattr(entry, '__iter__'):
                # entry is a sequence
                parts = [_ for _ in entry]
                if len(parts) > 1:
                    # take first two parts and skips others
                    new_vocab.append(entry[0:2])
                else:
                    # rare case, inner has one value only
                    new_vocab.append((entry[0], entry[0]))
        return new_vocab
    return definition


class Tag(object):

    def __init__(self, translate):
        self.translate = translate
        self.encoding = 'utf-8'

    def __call__(self, tag_name, *inners, **attributes):
        """Generates some xml/html tag.

        ``tagname``
            name of a valid tag

        ``inners``
            inner content of the tag. If empty a closed tag is generated

        ``attributes``
            attributes of the tag, leading or trailing ``_`` underscores are
            omitted from keywords.

        Example::

            >>> tag('p', 'Lorem Ipsum.', u'Hello World!',
            ...     class_='fancy', id='2f5b8a234ff')
            <p class="fancy" id="2f5b8a234ff">Lorem Ipsum. Hello World.</p>

        """
        cl = list()
        for key, value in attributes.items():
            if value is None or value is UNSET:
                continue
            value = self.translate(value)
            if not isinstance(value, UNICODE_TYPE):
                if isinstance(value, bytes):
                    value = value.decode(self.encoding)
                else:
                    value = str(value)
            cl.append((key.strip('_'), value))
        attributes = u''
        # NOTE: data attributes are enclosed in single quotes, since this makes
        # passing json lists possible. jQuery only recognizes JSON lists in
        # data attributes as such, if they are enclosed in single quotes,
        # because the JSON standard requires string values to be enclosed in
        # double quotes.
        if cl:
            attributes = list()
            for attr in cl:
                if 'data-' in attr[0]:
                    attributes.append(u"{0}='{1}'".format(*attr))
                else:
                    attributes.append(u'{0}="{1}"'.format(*attr))
            attributes = u' {0}'.format(u' '.join(sorted(attributes)))
        cl = list()
        for inner in inners:
            inner = self.translate(inner)
            if not isinstance(inner, UNICODE_TYPE):
                # XXX: inner = str(inner).decode(self.encoding)
                if isinstance(inner, bytes):
                    inner = inner.decode(self.encoding)
                else:
                    inner = str(inner)
            cl.append(inner)
        if not cl:
            return u'<{name}{attrs} />'.format(**{
                'name': tag_name,
                'attrs': attributes,
            })
        return u'<{name}{attrs}>{value}</{name}>'.format(**{
            'name': tag_name,
            'attrs': attributes,
            'value': u''.join(i for i in cl),
        })


# Deprecation message
def _deprecated_null_localization(msg):
    logging.warning(
        "Deprecated usage of 'yafowil.utils.tag', please "
        "use the tag factory on RuntimeData instead."
    )
    return msg


tag = Tag(_deprecated_null_localization)


class managedprops(object):

    def __init__(self, *args):
        self.__yafowil_managed_props__ = args

    def __call__(self, func):
        func.__yafowil_managed_props__ = self.__yafowil_managed_props__
        return func


def cssid(widget, prefix, postfix=None):
    if widget.attrs.get('structural', False):
        return None
    path = widget.dottedpath.replace(u'.', u'-')
    cssid = u'{0}-{1}'.format(prefix, path)
    if postfix is not None:
        cssid = u'{0}-{1}'.format(cssid, postfix)
    return unicodedata.normalize('NFKD', cssid)\
        .encode('ASCII', 'ignore')\
        .replace(b' ', b'_').decode()


def callable_value(value, widget, data):
    """Call value if callable with widget and data as arguments and return
    the callables return value. If value not callable, return as is.
    """
    if not callable(value):
        return value
    return value(widget, data)


def attr_value(key, widget, data, default=None):
    """Return widget attribute value by key or default. If value is callable,
    it's return value is used.
    """
    return callable_value(
        widget.attrs.get(key, default),
        widget,
        data
    )


def as_data_attrs(data):
    """Convert either dict or list of (key, value) pairs into dict containing
    HTML5 data attributes.

    Keys gets prefixed with ``data-``, ``CamelCase`` gets converted
    to ``caml-case``.

    Values are ignored if ``None`` or ``UNSET``. If value is string, it's
    taken as is, otherwise it's assumed that value is list or dict and gets
    dumped as JSON string.

    :param data: Either dict or list of (key, value) pairs.
    :return: Dict containing HTML5 data attributes
    """
    data_attrs = {}
    # no data passed, return empty dict
    if not data:
        return data_attrs
    # expect dict if no list
    if not isinstance(data, list):
        data = data.items()
    for key, val in data:
        # check against None and UNSET separately to please coverage tests
        # rnix, 2014-04-30
        if val is None:
            continue
        if val is UNSET:
            continue
        # convert value to JSON dump if no string.
        if not isinstance(val, STR_TYPE):
            # also remove leading and trailing double quotes,
            # they are not needed for data-attributes
            val = json.dumps(val).strip('"')
        # replace camelCase with camel-case
        key = re.sub(r'([a-z])([A-Z])', r'\g<1>-\g<2>', key).lower()
        data_attrs['data-{0}'.format(key)] = val
    return data_attrs


# B/C: deprecate as of yafowil 2.4, remove in yafowil 3.0
generic_html5_attrs = as_data_attrs


def data_attrs_helper(widget, data, attrs):
    """Creates a dictionary of JSON encoded data-attributes from a list of
    attribute-keys, ready to inject to a tag-renderer as expanded keyword
    arguments.

    :param widget: The yafowil widget.
    :param data: The data object.
    :param attrs: A list of data-attributes-keys to be used to generate the
                  data attributes dictionary.
    :type attrs: list
    :returns: Dictionary with keys as data-attribute-names, prefixed with
              'data-' and values from the widget.
    :rtype: dictionary

    The items in the list are the keys of the attributes for the target tag.
    Each key is prepended with 'data-'. The values are fetched from properties
    set on the widget. If a value is None, it isn't set. Other values are JSON
    encoded, which includes strings, booleans, lists, dicts.

    .. note::
      For camelCase attribute names are automatically split on word boundaries
      and made lowercase (e.g. camel-case). Since jQuery 1.6, the keys are
      converted to camelCase again after getting them with .data().

    .. note::
      The Tag class encloses data-attribute values in single quotes, since the
      JSON standard requires strings to be enclosed in double-quotes. jQuery
      requires this or .data() can't create lists or arrays out of
      data-attribute values.
    """
    items = [(key, attr_value(key, widget, data)) for key in attrs]
    return as_data_attrs(items)


css_managed_props = [
    'class', 'class_add',
    'valid_class', 'valid_class_default',
    'error_class', 'error_class_default',
    'required_class', 'required_class_default',
]


def cssclasses(widget, data, classattr='class', additional=[]):
    _classes = list()
    attrs = widget.attrs
    if attrs['error_class'] and data.errors:
        if isinstance(attrs['error_class'], STR_TYPE):
            _classes.append(attrs['error_class'])
        else:
            _classes.append(attrs['error_class_default'])
    elif attrs['valid_class'] and data.root.extracted:
        if isinstance(attrs['valid_class'], STR_TYPE):
            _classes.append(attrs['valid_class'])
        else:
            _classes.append(attrs['valid_class_default'])
    if attrs['required_class'] and attrs['required']:
        if isinstance(attrs['required_class'], STR_TYPE):
            _classes.append(attrs['required_class'])
        else:
            _classes.append(attrs['required_class_default'])
    if attrs[classattr]:
        _classes += attr_value(classattr, widget, data).split()
    if attrs['class_add']:
        _classes += attr_value('class_add', widget, data).split()
    additional = [add for add in additional if add]
    _classes += additional
    return _classes and ' '.join(sorted(_classes)) or None


class EmptyValue(object):
    """Used to identify empty values in conjunction with data type conversion.
    """

    def __nonzero__(self):
        return False

    __bool__ = __nonzero__

    def __str__(self):
        return ''

    def __len__(self):
        return 0

    def __repr__(self):
        return '<EMPTY_VALUE>'


EMPTY_VALUE = EmptyValue()


DATATYPE_PRECONVERTERS = {
    float: lambda x: isinstance(x, STR_TYPE) and x.replace(',', '.') or x
}
# B/C
DATATYPE_PRECONVERTERS['float'] = DATATYPE_PRECONVERTERS[float]
DATATYPE_CONVERTERS = {
    'str': BYTES_TYPE,
    'unicode': UNICODE_TYPE,
    'int': int,
    'integer': int,
    'long': LONG_TYPE,
    'float': float,
    'uuid': uuid.UUID
}


def convert_value_to_datatype(value, datatype, empty_value=EMPTY_VALUE):
    """Convert given value to datatype.

    Datatype is either a callable or a string out of ``'str'``, ``'unicode'``,
    ``'int'``, ``'integer'``, ``'long'``, ``'float'`` or ``'uuid'``

    If value is ``UNSET``, return ``UNSET``, regardless of given datatype.

    If value is ``EMPTY_VALUE``, return ``empty_value``, which defaults to
    ``EMPTY_VALUE`` marker.

    If value is ``None`` or ``''``, return ``empty_value``, which defaults to
    ``EMPTY_VALUE`` marker. Be aware that empty value marker is even returned
    if ``str`` datatype, to provide a consistent behavior.

    Converter callables must raise one out of the following exceptions if
    conversion fails:

        * ``ValueError``
        * ``UnicodeDecodeError``
        * ``UnicodeEncodeError``
    """
    if value is UNSET:
        return UNSET
    if value is EMPTY_VALUE:
        return empty_value
    if value in [None, '']:
        return empty_value
    if isinstance(datatype, STR_TYPE):
        converter = DATATYPE_CONVERTERS[datatype]
    else:
        converter = datatype
    try:
        if isinstance(value, converter):
            return value
    except TypeError:
        # converter is instance of class or function
        pass
    preconverter = DATATYPE_PRECONVERTERS.get(datatype)
    if preconverter:
        value = preconverter(value)
    # special case bytes or str buildin type in python 3
    # uses ascii codec to emulate same behavior as when converting with python2
    # this is supposed to change in future
    if not IS_PY2 and converter in (bytes, str):
        return converter(value, 'ascii')  # pragma: no cover
    return converter(value)


def convert_values_to_datatype(value, datatype, empty_value=EMPTY_VALUE):
    if isinstance(value, list):
        res = list()
        for item in value:
            res.append(convert_value_to_datatype(
                item,
                datatype,
                empty_value=empty_value
            ))
        return res
    return convert_value_to_datatype(value, datatype, empty_value=empty_value)
